from __future__ import annotations

from prettyqt import widgets


class SingleLineTextEdit(widgets.PlainTextEdit):
    def __init__(self, *args, object_name: str = "singleline_textedit", **kwargs):
        super().__init__(*args, object_name=object_name, **kwargs)
        self.textChanged.connect(self._on_text_changed)
        font_metrics = self.get_font_metrics()
        row_height = font_metrics.lineSpacing()
        self.setFixedHeight(int(row_height * 1.5))
        self.set_size_policy(vertical="fixed")
        self.set_line_wrap_mode("none")
        self.set_scrollbar_policy("always_off")

    def _on_text_changed(self):
        text = self.text()
        with self.selecter.current_cursor() as c:
            pos = c.position()
            num_linebreaks = text.count("\n")
            with self.signals_blocked():
                self.set_text(text.replace("\n", ""))
            c.setPosition(pos - num_linebreaks)


if __name__ == "__main__":
    app = widgets.app()
    widget = SingleLineTextEdit()
    widget.show()
    app.exec()



# modified version of https://github.com/nicoddemus/qt-async-threads/

from __future__ import annotations

from collections.abc import (
    Awaitable,
)
from concurrent.futures import Future, ThreadPoolExecutor
from contextlib import suppress
import dataclasses
import functools
import itertools
import logging
import os
import threading
import time
from typing import TYPE_CHECKING, Any, Self, cast

from prettyqt import core


if TYPE_CHECKING:
    from collections.abc import (
        AsyncIterator,
        Callable,
        Coroutine,
        Generator,
        Iterable,
        Iterator,
    )

    from requests import Response


log = logging.getLogger(__name__)


class AsyncRunner:
    """A runner which runs long-lasting functions using a thread pool."""

    def __init__(self, max_threads: int | None = None):
        super().__init__()
        self._max_threads = max_threads or os.cpu_count() or 1
        self._pool = ThreadPoolExecutor(max_workers=max_threads)

        # Keep track of running tasks,
        # mostly to know if we are idle or not.
        self._running_tasks: list[_AsyncTask] = []

        # This signaller object is used to signal to us when a future running
        # in another thread finishes. Thanks to Qt's queued connections,
        # signals can be safely emitted from separate threads and are queued
        # to run in the same thread as the object receiving it lives (the main
        # thread in our case).
        self._signaller = _FutureDoneSignaller()
        self._signaller.future_done_signal.connect(self._resume_coroutine)

    def __enter__(self) -> Self:
        return self

    def __exit__(self, *exc_info: object):
        self.close()

    def to_sync(
        self, async_func: Callable[..., Coroutine[Any, Any, None]]
    ) -> Callable[..., None]:
        """Return new sync function.

        Return a new sync function that will start its coroutine using
        ``start_coroutine`` when called, returning immediately.
        Use to connect Qt signals to async functions

        Arguments:
            async_func: Async Qt Slot

        Returns:
            new sync function

        ### Example:
        ```py
        dl_btn.clicked.connect(self.runner.to_sync(self._on_dl_btn_clicked)
        ...
        search_response = await self.runner.run(requests.get, api_url)
        ```
        """

        @functools.wraps(async_func)
        def func(*args: object, **kwargs: object):
            gen = async_func(*args, **kwargs)
            self.start_coroutine(gen)

        return func

    @property
    def max_threads(self) -> int:
        """Return the maximum number of threads used by the internal threading pool."""
        return self._max_threads

    def is_idle(self) -> bool:
        return len(self._running_tasks) == 0

    def close(self):
        self._pool.shutdown(wait=True, cancel_futures=True)

    async def run[**P, T](
        self, func: Callable[P, T], *args: P.args, **kwargs: P.kwargs
    ) -> T | None:
        """Run the given function in a thread.

        While it is running, yields
        the control back to the Qt event loop.

        When the thread finishes, this async function resumes, returning
        the return value from the function.

        Arguments:
            func: Callable to run in a thrad
            args: arguments passed to Callable
            kwargs: keyword arguments passed to Callable

        Returns:
            return value from function
        """
        funcs = [functools.partial(func, *args, **kwargs)]
        async for result in self.run_parallel(funcs):
            return result
        return None

    async def run_parallel[T](  # type:ignore[override]
        self, funcs: Iterable[Callable[[], T]]
    ) -> AsyncIterator[T]:
        """Run functions in parallel.

        Run functions in parallel without arguments, yielding results as they get ready.
        (use ``functools.partial`` as necessary)

        Arguments:
            funcs: Sequence of Functions to call.
        """
        # We submit the functions in batches to avoid overloading the pool,
        # which can cause other coroutines to stall.
        # For example, Autofit works by executing 1000s of small functions,
        # which might take a few ms each. If we submitted all those 1000s of functions
        # at once (at the time ``run_parallel`` is called), then other coroutines
        # that try to submit functions to execute in threads would only be resumed
        # much later, causing a noticeable slow down in the application.
        batch_size = max(self._max_threads // 2, 1)
        for function_batch in itertools.batched(funcs, batch_size):
            # Submit all functions from the current batch to the thread pool,
            # using the _AsyncTask to track the futures and await when they finish.
            task = _AsyncTask({self._pool.submit(f) for f in function_batch})
            self._running_tasks.append(task)
            for future in task.futures:
                future.add_done_callback(
                    functools.partial(self._on_task_future_done, task=task)
                )
            try:
                # We wait until all functions in this batch have finished
                # until we submit the next batch. This is not 100% optimal
                # but ensures we get a fairer share of the pool.
                while task.futures:
                    await task
                    for result in task.pop_done_futures():
                        yield result
            finally:
                task.shutdown()
                self._running_tasks.remove(task)

    def _on_future_done(
        self,
        future: Future,
        *,
        coroutine: Coroutine[Any, Any, Any],
    ):
        """Called when a ``Future`` that was submitted to the thread pool finishes.

        This function is called from a separate thread, so we emit the signal
        of the internal ``_signaller``, which thanks to Qt's queued connections feature,
        will post the event to the main loop, and it will be processed there.

        Arguments:
            future: future which finished
            coroutine: Coroutine
        """
        self._signaller.future_done_signal.emit(future, coroutine)

    def _on_task_future_done(self, future: Future, *, task: _AsyncTask):
        """Called when a ``Future`` belonging to a ``_AsyncTask`` has finished.

        Similar to ``_on_future_done``, this emits a signal so the coroutine is resumed
        in the main thread.

        Arguments:
            future: future which finished
            task:  Async Task
        """
        # At this point, we want to get the coroutine associated with the task, and resume
        # its execution in the main loop, but we must take care bc this callback is called
        # from another thread, and it may be called multiple times in succession, but we
        # want only **one** event to be sent, that's
        # why we use ``pop_coroutine``, which is lock-protected and will return
        # the coroutine and set it to None, so next calls of this method will get ``None``
        # and won't trigger the event again.
        if coroutine := task.pop_coroutine():
            self._signaller.future_done_signal.emit(future, coroutine)

    def _resume_coroutine(
        self,
        future: Future,
        coroutine: Coroutine[Any, Any, Any],
    ):
        """Resume paused coroutine.

        Slots connected to our internal ``_signaller`` object,
        called in the main thread after a future finishes, resuming the paused coroutine.

        Arguments:
            future: not used
            coroutine: coroutine to resume
        """
        assert threading.current_thread() is threading.main_thread()
        self.start_coroutine(coroutine)

    def start_coroutine(self, coroutine: Coroutine):
        """Start the coroutine, and returns immediately.

        Arguments:
            coroutine: coroutine to start
        """
        # Note: this function will also be called to resume a paused coroutine that was
        # waiting for a thread to finish (by ``_resume_coroutine``).
        with suppress(StopIteration):
            value = coroutine.send(None)
            if isinstance(value, _AsyncTask):
                # At this point, return control to the event loop; when one
                # of the futures running in the task finishes, it will resume the
                # coroutine back in the main thread.
                value.coroutine = coroutine
                return
            raise ValueError(value)

    def run_coroutine[T](self, coroutine: Coroutine[Any, Any, T]) -> T:
        """Start the coroutine.

        Starts the coroutine, doing a busy loop while waiting for it to complete,
        returning then the result.

        Arguments:
            coroutine: coroutine to start

        Note: see warning in AbstractAsyncRunner about when to use this function.
        """
        result: T | None = None
        exception: Exception | None = None
        completed = False

        async def wrapper():
            nonlocal result, exception, completed
            try:
                result = await coroutine
            except Exception as e:  # noqa: BLE001
                exception = e
            completed = True

        self.start_coroutine(wrapper())
        while not completed:
            core.CoreApplication.processEvents()
            time.sleep(0.01)

        if exception is not None:
            raise exception
        return cast(T, result)


@dataclasses.dataclass
class _AsyncTask(Awaitable[None]):
    """Awaitable that is propagated up the async stack, containing running futures.

    It "awaits" while all futures are processing, and will
    stop and return once any of them complete.
    """

    futures: set[Future]
    coroutine: Coroutine | None = None
    _lock: threading.Lock = dataclasses.field(default_factory=threading.Lock)

    def __await__(self) -> Generator[Any, None, None]:
        if any(x for x in self.futures if x.done()):
            return None
        else:
            yield self

    def pop_done_futures(self) -> Iterator[Any]:
        """Yield done futures, removing them from the ``futures`` simultaniously."""
        for future in list(self.futures):
            if future.done():
                self.futures.discard(future)
                yield future.result()

    def pop_coroutine(self) -> Coroutine | None:
        """Return current coroutine.

        Returns the current coroutine associated with this object, while
        also setting the ``coroutine`` attribute to ``None``. This
        is meant to be called from multiple threads, in a way that only
        one of them will be able to obtain the coroutine object, while the
        others will get ``None``.
        """
        with self._lock:
            coroutine = self.coroutine
            self.coroutine = None
            return coroutine

    def shutdown(self):
        """Cancel any running futures and clears up its attributes."""
        msg = (
            "Should always be called from ``run_parallel``, "
            "at which point this should not have a coroutine associated"
        )
        assert self.coroutine is None, msg
        while self.futures:
            future = self.futures.pop()
            future.cancel()


class _FutureDoneSignaller(core.Object):
    """Used to emit a signal to the main thread when a future finishes in another."""

    # This emits(Future, Coroutine). We need to use ``object`` as the
    # second parameter because Qt doesn't allow us to use an ABC class there it seems.
    future_done_signal = core.Signal(Future, object)


if __name__ == "__main__":
    from pathlib import Path
    from urllib.parse import urlsplit

    import requests
    from requests.exceptions import ConnectionError as ConnError

    from prettyqt import widgets

    class Window(widgets.Widget):
        def __init__(self, directory: Path):
            super().__init__()

            self.setWindowTitle("Cat Downloader")
            self.directory = directory
            self._cancelled = False

            # Build controls.
            self.count_spin = widgets.SpinBox(value=5, minimum=1)
            self.progress_label = widgets.Label("Idle, click below to start downloading")
            self.download_button = widgets.PushButton(
                "Download", clicked=self.on_download_button_clicked
            )
            self.stop_button = widgets.PushButton(
                "Stop", enabled=False, clicked=self.on_cancel_button_clicked
            )

            layout = widgets.FormLayout(self)
            layout.addRow("How many cats?", self.count_spin)
            layout.addRow("Status", self.progress_label)
            layout.addRow(self.download_button)
            layout.addRow(self.stop_button)

        async def on_download_button_clicked(self, checked: bool = False):
            self.progress_label.setText("Searching...")
            self.download_button.setEnabled(False)
            self.stop_button.setEnabled(True)

            self._cancelled = False
            downloaded_count = 0

            def download_one() -> Response:
                # Search.
                search_response = requests.get(
                    "https://api.thecatapi.com/v1/images/search"
                )
                search_response.raise_for_status()

                # Download.
                url = search_response.json()[0]["url"]
                return requests.get(url)

            try:
                functions = [download_one for _ in range(self.count_spin.value())]
                async for response in widgets.app().run_parallel(functions):
                    # Save the contents of the image to a file.
                    parts = urlsplit(response.url)
                    path = (
                        self.directory
                        / f"{downloaded_count:02d}_cat{Path(parts.path).suffix}"
                    )
                    path.write_bytes(response.content)
                    downloaded_count += 1

                    # Show progress.
                    self.progress_label.setText(f"Downloaded {path.name}")
                    widgets.app().processEvents()
                    if self._cancelled:
                        widgets.MessageBox.information(
                            self, "Cancelled", "Download cancelled"
                        )
                        return
            except ConnError as e:
                widgets.MessageBox.critical(self, "Error", e)
                return
            finally:
                self.progress_label.setText(f"Done, {downloaded_count} cats downloaded")
                self.download_button.setEnabled(True)
                self.stop_button.setEnabled(False)

        def on_cancel_button_clicked(self):
            self._cancelled = True

    app = widgets.app()
    win = Window(Path(__file__).parent)
    win.show()
    app.exec()



from __future__ import annotations

from prettyqt import constants, core, itemmodels


class Root:
    def __init__(self, name="Root", children=None):
        self.children = children or []
        self.name = name


class NestedModel(  # type: ignore
    itemmodels.ListMixin, itemmodels.ModelMixin, core.AbstractItemModel
):
    DEFAULT_FLAGS = (
        constants.DRAG_ENABLED  # type: ignore
        | constants.IS_ENABLED
        | constants.IS_SELECTABLE
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.root = Root()
        self.items = self.root.children

    def flags(self, index: core.ModelIndex) -> constants.ItemFlag:
        if not index.isValid():
            return constants.NO_FLAGS
        if index.column() in self.SET_DATA:
            return self.DEFAULT_FLAGS | constants.IS_EDITABLE
        return self.DEFAULT_FLAGS

    def rowCount(self, parent: core.ModelIndex | None = None) -> int:
        parent = parent or core.ModelIndex()
        if parent.column() > 0:
            return 0
        return (
            len(parent.internalPointer().children)
            if parent.isValid()
            else len(self.items)
        )

    def index(
        self, row: int, column: int, parent: core.ModelIndex | None = None
    ) -> core.ModelIndex:
        parent = parent or core.ModelIndex()
        if not self.hasIndex(row, column, parent):
            return core.ModelIndex()

        parent_item = parent.internalPointer()
        if not parent_item:
            parent_item = self.root

        return self.createIndex(row, column, parent_item.children[row])

    def parent(self, index: core.ModelIndex | None = None) -> core.ModelIndex:
        # hacky way to let the case without any arguments get through.
        # not really nice, a proper dispatch library would be better.
        # functools.singledispatchmethod doesnt work here.
        if index is None:
            return super().parent()
        if not index.isValid():
            return core.ModelIndex()

        if item := index.internalPointer():
            return (
                core.ModelIndex()
                if item.parent in [self.root, None]
                else self.createIndex(item.parent.row(), 0, item.parent)
            )
        return core.ModelIndex()

    def data_by_index(self, index):
        return index.internalPointer()

    def json(self, root=None) -> dict:
        """Serialise model as JSON-compliant dictionary.

        Arguments:
            root (JsonTreeItem, optional): Serialise from here
                defaults to the the top-level item

        Returns:
            model as dict
        """
        root = root or self.root
        return root.as_json()



from __future__ import annotations

from prettyqt import positioning
from prettyqt.qt import QtCore, QtPositioning
from prettyqt.utils import bidict, get_repr


QNmeaPositionInfoSource = QtPositioning.QNmeaPositionInfoSource

UPDATE_MODES = bidict(
    real_time=QtPositioning.QNmeaPositionInfoSource.UpdateMode.RealTimeMode,
    simulation=QtPositioning.QNmeaPositionInfoSource.UpdateMode.SimulationMode,
)


class NmeaPositionInfoSource(
    positioning.GeoPositionInfoSourceMixin, QtPositioning.QNmeaPositionInfoSource
):
    def __init__(
        self,
        update_mode: QtPositioning.QNmeaPositionInfoSource.UpdateMode | str,
        parent: QtCore.QObject,
    ):
        if isinstance(update_mode, QtPositioning.QNmeaPositionInfoSource.UpdateMode):
            mode = update_mode
        else:
            mode = UPDATE_MODES[update_mode]
        super().__init__(mode, parent)

    def __repr__(self):
        return get_repr(self)

    def get_update_mode(self) -> str:
        return UPDATE_MODES.inverse[self.updateMode()]


if __name__ == "__main__":
    obj = QtCore.QObject()
    source = NmeaPositionInfoSource("real_time", obj)



from __future__ import annotations

from dataclasses import dataclass
from typing import Any

from prettyqt import custom_widgets, widgets


@dataclass
class BaseSetting:
    identifier: str
    label: str
    description: str
    requires_restart: bool = False


@dataclass
class SelectionSetting(BaseSetting):
    options: list[str] | None = None
    default: Any = None
    minimum: float = 0.0
    maximum: float = 0.0
    pattern: str = ""
    requires_restart: bool = False


@dataclass
class BoolSetting(BaseSetting):
    default: bool = False


@dataclass
class IntSetting(BaseSetting):
    options: list[str] | None = None
    default: Any = None
    minimum: int | None = None
    maximum: int | None = None


@dataclass
class StrSetting(BaseSetting):
    default: Any = None
    minimum: int | None = None
    maximum: int | None = None


class SettingsWindow(widgets.Widget):
    """Widget to easilty create settings widgets based on dataclasses."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        layout = self.set_layout("horizontal")
        self.configwidget = widgets.Widget()
        self.configwidget.set_max_width(600)
        self.configwidget.set_layout("vertical")
        self.scrollarea = widgets.ScrollArea(widget_resizable=True)
        self.scrollarea.set_widget(self.configwidget)
        self.toc = custom_widgets.ScrollAreaTocWidget(self.scrollarea)
        right_layout = widgets.VBoxLayout()
        self.search_lineedit = widgets.LineEdit()
        self.search_lineedit.set_max_width(200)
        self.search_lineedit.setPlaceholderText("Search setting...")
        self.search_lineedit.value_changed.connect(self._on_search)
        right_layout.add(self.search_lineedit)
        right_layout.add(self.scrollarea)
        layout.add(self.toc)
        layout.add(right_layout)

    def _on_search(self, search_term: str):
        for widget in self.scrollarea.find_children(widgets.GroupBox):
            widget.setVisible(search_term.lower() in widget.windowTitle().lower())
            # TODO: this doesnt work correctly yet.
            # self.toc.proxy.set_search_term(search_term)

    def add_setting(self, setting: BaseSetting):
        row_widget = widgets.GroupBox(window_title=setting.label)
        layout = row_widget.set_layout("vertical")
        header_label = widgets.Label(setting.label)
        with header_label.edit_font() as font:
            font.setPointSize(14)
        layout.add(header_label)
        if setting.requires_restart:
            label = widgets.Label("(requires restart)")
            label.set_color("red")
            layout.add(label)
        match setting:
            case SelectionSetting():
                widget = widgets.ComboBox()
                widget.add_items(setting.options)
            case BoolSetting():
                widget = widgets.CheckBox()
            case StrSetting():
                widget = widgets.LineEdit()
            case IntSetting():
                widget = widgets.SpinBox()
                widget.set_range(setting.minimum, setting.maximum)

        container = widgets.Widget()
        row_layout = container.set_layout("horizontal")
        row_layout.add(widget)
        row_layout.add(widgets.Label(setting.description), stretch=1)
        if setting.default is not None:
            widget.set_value(setting.default)
        layout.add(container)
        self.configwidget.box.add(row_widget)


if __name__ == "__main__":
    app_style = SelectionSetting(
        identifier="app_style",
        label="App style",
        description="Set the app style",
        options=widgets.StyleFactory.keys(),
        requires_restart=True,
        default=None,
    )

    bool_test = BoolSetting(
        identifier="bool_setting",
        label="Bool setting",
        description="This setting is a bool.",
        default=True,
    )
    int_test = IntSetting(
        identifier="int_setting",
        label="Int setting",
        description="This setting is an int.",
        default=15,
        minimum=10,
        maximum=20,
    )

    str_test = StrSetting(
        identifier="str_setting",
        label="Str setting",
        description="Some longer text",
        requires_restart=True,
        default="hallo",
    )

    app = widgets.app()
    widget = SettingsWindow()
    widget.add_setting(app_style)
    widget.add_setting(bool_test)
    widget.add_setting(bool_test)
    widget.add_setting(bool_test)
    widget.add_setting(str_test)
    widget.add_setting(int_test)
    widget.add_setting(str_test)
    widget.add_setting(str_test)
    widget.add_setting(app_style)
    widget.add_setting(app_style)
    widget.add_setting(app_style)
    widget.add_setting(app_style)
    widget.show()
    app.exec()



from __future__ import annotations

from prettyqt import core, gui
from prettyqt.utils import get_repr


class DoubleValidator(gui.ValidatorMixin, gui.QDoubleValidator):
    """Range checking of floating-point numbers."""

    ID = "double"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setLocale(core.Locale("en-En"))

    def __repr__(self):
        return get_repr(self, self.bottom(), self.top(), self.decimals())

    def __reduce__(self):
        return type(self), (self.bottom(), self.top(), self.decimals()), None

    def __eq__(self, other: object):
        return (
            (
                self.bottom() == other.bottom()
                and self.top() == other.top()
                and self.decimals() == other.decimals()
            )
            if isinstance(other, type(self))
            else False
        )

    def set_range(
        self, start: float | None, end: float | None, decimals: int | None = None
    ):
        if decimals is None:
            decimals = -1
        if start is None:
            start = -float("inf")
        if end is None:
            end = float("inf")
        self.setRange(start, end, decimals)


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    lineedit = widgets.LineEdit()
    val = DoubleValidator()
    # val.setRange(0, 9)
    lineedit.set_validator(val)
    lineedit.show()
    app.exec()



from __future__ import annotations

from prettyqt import constants, widgets


class GraphicsAnchorLayout(widgets.GraphicsLayoutMixin, widgets.QGraphicsAnchorLayout):
    def add_anchor(
        self,
        first_item: widgets.QGraphicsLayoutItem,
        first_edge: constants.AnchorPointStr,
        second_item: widgets.QGraphicsLayoutItem,
        second_edge: constants.AnchorPointStr,
    ) -> widgets.QGraphicsAnchor:
        return self.addAnchor(
            first_item,
            constants.ANCHOR_POINT[first_edge],
            second_item,
            constants.ANCHOR_POINT[second_edge],
        )

    def get_anchor(
        self,
        first_item: widgets.QGraphicsLayoutItem,
        first_edge: constants.AnchorPointStr,
        second_item: widgets.QGraphicsLayoutItem,
        second_edge: constants.AnchorPointStr,
    ) -> widgets.QGraphicsAnchor:
        return self.anchor(
            first_item,
            constants.ANCHOR_POINT[first_edge],
            second_item,
            constants.ANCHOR_POINT[second_edge],
        )

    def add_anchors(
        self,
        first_item: widgets.QGraphicsLayoutItem,
        second_item: widgets.QGraphicsLayoutItem,
        orientation: constants.OrientationStr | constants.Orientation,
    ):
        self.addAnchors(
            first_item, second_item, constants.ORIENTATION.get_enum_value(orientation)
        )

    def add_corner_anchors(
        self,
        first_item: widgets.QGraphicsLayoutItem,
        first_corner: constants.CornerStr,
        second_item: widgets.QGraphicsLayoutItem,
        second_corner: constants.CornerStr,
    ):
        self.addCornerAnchors(
            first_item,
            constants.CORNER[first_corner],
            second_item,
            constants.CORNER[second_corner],
        )


if __name__ == "__main__":
    app = widgets.app()
    layout = GraphicsAnchorLayout()



from __future__ import annotations

from typing import Literal

from prettyqt import core
from prettyqt.qt import QtNetwork
from prettyqt.utils import bidict, datatypes


ContentTypeStr = Literal["mixed", "related", "form", "alternative"]

CONTENT_TYPES: bidict[ContentTypeStr, QtNetwork.QHttpMultiPart.ContentType] = bidict(
    mixed=QtNetwork.QHttpMultiPart.ContentType.MixedType,
    related=QtNetwork.QHttpMultiPart.ContentType.RelatedType,
    form=QtNetwork.QHttpMultiPart.ContentType.FormDataType,
    alternative=QtNetwork.QHttpMultiPart.ContentType.AlternativeType,
)


class HttpMultiPart(core.ObjectMixin, QtNetwork.QHttpMultiPart):
    """Resembles a MIME multipart message to be sent over HTTP."""

    def __add__(self, other: QtNetwork.QHttpPart):
        self.append(other)
        return self

    def set_content_type(
        self, typ: ContentTypeStr | QtNetwork.QHttpMultiPart.ContentType
    ):
        """Set content type.

        Args:
            typ: content type
        """
        self.setContentType(CONTENT_TYPES.get_enum_value(typ))

    def set_boundary(self, boundary: datatypes.ByteArrayType):
        boundary = datatypes.to_bytearray(boundary)
        self.setBoundary(boundary)

    def get_boundary(self) -> str:
        return self.boundary().data().decode()


if __name__ == "__main__":
    part = HttpMultiPart()
    print(type(part.boundary().data()))



from __future__ import annotations

from prettyqt.qt import QtQml


class JSValueIterator(QtQml.QJSValueIterator):
    """Java-style iterator for JSValue."""

    def __iter__(self):
        return self

    def __next__(self):
        if self.next():
            return (self.name(), self.value().toVariant())
        raise StopIteration



"""Provides QtTest classes and functions."""

from prettyqt.qt import PYSIDE6


if PYSIDE6:
    from PySide6.QtUiTools import QUiLoader, loadUiType  # type: ignore

    RCC_CMD = "pyside6-rcc --no-compress --verbose"
    LUPDATE_CMD = "pyside6-lupdate -verbose"
    UIC_CMD = "pyside6-uic"
else:
    raise ModuleNotFoundError("No Qt bindings could be found")



from __future__ import annotations

from typing import Literal

from prettyqt.qt import QtWidgets
from prettyqt.utils import bidict


mod = QtWidgets.QTreeWidgetItemIterator

IteratorFlagStr = Literal[
    "all",
    "hidden",
    "not_hidden",
    "selected",
    "unselected",
    "selectable",
    "not_selectable",
    "drag_enabled",
    "drag_disabled",
    "drop_enabled",
    "drop_disabled",
    "has_children",
    "no_children",
    "checked",
    "not_checked",
    "enabled",
    "disabled",
    "editable",
    "not_editable",
    "user_flag",
]

ITERATOR_FLAGS = bidict(
    all=mod.IteratorFlag.All,
    hidden=mod.IteratorFlag.Hidden,
    not_hidden=mod.IteratorFlag.NotHidden,
    selected=mod.IteratorFlag.Selected,
    unselected=mod.IteratorFlag.Unselected,
    selectable=mod.IteratorFlag.Selectable,
    not_selectable=mod.IteratorFlag.NotSelectable,
    drag_enabled=mod.IteratorFlag.DragEnabled,
    drag_disabled=mod.IteratorFlag.DragDisabled,
    drop_enabled=mod.IteratorFlag.DropEnabled,
    drop_disabled=mod.IteratorFlag.DropDisabled,
    has_children=mod.IteratorFlag.HasChildren,
    no_children=mod.IteratorFlag.NoChildren,
    checked=mod.IteratorFlag.Checked,
    not_checked=mod.IteratorFlag.NotChecked,
    enabled=mod.IteratorFlag.Enabled,
    disabled=mod.IteratorFlag.Disabled,
    editable=mod.IteratorFlag.Editable,
    not_editable=mod.IteratorFlag.NotEditable,
    user_flag=mod.IteratorFlag.UserFlag,
)


class TreeWidgetItemIterator(QtWidgets.QTreeWidgetItemIterator):
    def __init__(
        self,
        other: (
            QtWidgets.QTreeWidget
            | QtWidgets.QTreeWidgetItem
            | QtWidgets.QTreeWidgetItemIterator
        ),
        flags: QtWidgets.QTreeWidgetItemIterator.IteratorFlag | None = None,
        hidden: bool | None = None,
        selected: bool | None = None,
        selectable: bool | None = None,
        draggable: bool | None = None,
        droppable: bool | None = None,
        has_children: bool | None = None,
        checked: bool | None = None,
        enabled: bool | None = None,
        editable: bool | None = None,
        user_flag: bool = False,
    ):
        if isinstance(other, QtWidgets.QTreeWidgetItemIterator):
            super().__init__(other)
        else:
            if flags is None:
                flags = mod.IteratorFlag.All  # type: ignore
            if hidden is True:
                flags |= mod.IteratorFlag.Hidden  # type: ignore
            elif hidden is False:
                flags |= mod.IteratorFlag.NotHidden  # type: ignore
            if selected is True:
                flags |= mod.IteratorFlag.Selected  # type: ignore
            elif selected is False:
                flags |= mod.IteratorFlag.Unselected  # type: ignore
            if selectable is True:
                flags |= mod.IteratorFlag.Selectable  # type: ignore
            elif selectable is False:
                flags |= mod.IteratorFlag.NotSelectable  # type: ignore
            if draggable is True:
                flags |= mod.IteratorFlag.DragEnabled  # type: ignore
            elif draggable is False:
                flags |= mod.IteratorFlag.DragDisabled  # type: ignore
            if droppable is True:
                flags |= mod.IteratorFlag.DropEnabled  # type: ignore
            elif droppable is False:
                flags |= mod.IteratorFlag.DropDisabled  # type: ignore
            if has_children is True:
                flags |= mod.IteratorFlag.HasChildren  # type: ignore
            elif has_children is False:
                flags |= mod.IteratorFlag.NoChildren  # type: ignore
            if checked is True:
                flags |= mod.IteratorFlag.Checked  # type: ignore
            elif checked is False:
                flags |= mod.IteratorFlag.NotChecked  # type: ignore
            if enabled is True:
                flags |= mod.IteratorFlag.Enabled  # type: ignore
            elif enabled is False:
                flags |= mod.IteratorFlag.Disabled  # type: ignore
            if editable is True:
                flags |= mod.IteratorFlag.Editable  # type: ignore
            elif editable is False:
                flags |= mod.IteratorFlag.NotEditable  # type: ignore
            if user_flag:
                flags |= mod.IteratorFlag.UserFlag  # type: ignore
            super().__init__(other, flags)  # type: ignore


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    item = QtWidgets.QTreeWidget()
    widget = TreeWidgetItemIterator(item)



# This is a PyInstaller `hook  # noqa: N999
# <https://pyinstaller.readthedocs.io/en/stable/hooks.html>`_.
# See the `PyInstaller manual <https://pyinstaller.readthedocs.io/>`_
# for more information.
#
from __future__ import annotations

from PyInstaller.utils.hooks import collect_data_files


# For more information see
# `hook global variables
# <https://pyinstaller.readthedocs.io/en/stable/hooks.html#hook-global-variables>`_
# in the manual for more information.

# hiddenimports = ["pyi_hooksample._hidden"]
# The ``excludes`` parameter of `collect_data_files
# <https://pyinstaller.readthedocs.io/en/stable/hooks.html#useful-items-in-pyinstaller-utils-hooks>`_
# excludes ``rthooks.dat`` from the frozen executable, which is only needed when
# freezeing, but not when executing the frozen program.
datas = collect_data_files("prettyqt", excludes=["__pyinstaller"])



from __future__ import annotations

from typing import TYPE_CHECKING

from prettyqt import iconprovider, quick


if TYPE_CHECKING:
    from prettyqt.qt import QtCore


class AwesomeQuickImageProvider(quick.QuickImageProvider):
    def requestPixmap(self, id_: str, requested_size: QtCore.QSize):
        pix = iconprovider.get_icon(id_).pixmap(requested_size)
        return pix, pix.size()

    def requestImage(self, id_: str, requested_size: QtCore.QSize):
        img = iconprovider.get_icon(id_).pixmap(requested_size).toImage()
        return img, img.size()



from __future__ import annotations

from prettyqt import constants, core, widgets


class OrientedScrollArea(widgets.ScrollArea):
    """ScrollArea with helper methods if the Area is treated as having an orientation."""

    def __init__(
        self,
        orientation: constants.Orientation = constants.Orientation.Horizontal,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self._orientation = orientation
        self.set_orientation(orientation)

    def orientation(self) -> constants.Orientation:
        """Get the orientation."""
        return self._orientation

    def set_orientation(self, orientation: constants.Orientation):
        """Set the orientation."""
        self._orientation = orientation
        if orientation == constants.Orientation.Horizontal:
            self.set_horizontal_scrollbar_policy("always_on")
            self.set_vertical_scrollbar_policy("always_off")
            self.set_size_policy("minimum_expanding", "fixed")
        else:
            self.set_horizontal_scrollbar_policy("always_off")
            self.set_vertical_scrollbar_policy("always_on")
            self.set_size_policy("fixed", "minimum_expanding")
        self.update()

    def eventFilter(self, obj: core.QObject, event: core.QEvent) -> bool:
        """Filter events in order to get informed when the content widget resizes."""
        if obj == self.widget() and event.type() == core.QEvent.Type.Resize:
            self.updateGeometry()
        return super().eventFilter(obj, event)

    def sizeHint(self) -> core.QSize:
        """Determines the exact size along the axis orthogonal to the orientation."""
        widget = self.widget()
        if widget is None:
            return super().sizeHint()
        widget_size = widget.size()
        margins = self.contentsMargins()
        margins_width = margins.left() + margins.right()
        margins_height = margins.top() + margins.bottom()
        w = widget_size.width() + margins_width
        h = widget_size.height() + margins_height
        if self._orientation == constants.Orientation.Horizontal:
            return core.QSize(w, h + self.horizontalScrollBar().sizeHint().height())
        # self._orientation == Qt.Vertical:
        return core.QSize(w + self.verticalScrollBar().sizeHint().width(), h)


if __name__ == "__main__":
    app = widgets.app()
    area = OrientedScrollArea()
    w = widgets.PlainTextEdit()
    # w.set_layout("horizontal")
    area.setWidget(w)
    # for i in range(100):
    #     w.box.addWidget(widgets.PlainTextEdit("test"))
    area.show()
    with app.debug_mode():
        app.exec()



from __future__ import annotations

import ast
import logging
from typing import TYPE_CHECKING

from prettyqt import constants, custom_widgets, itemmodels, widgets


if TYPE_CHECKING:
    from prettyqt import core


logger = logging.getLogger(__name__)


class AstViewer(widgets.Splitter):
    """Widget which displays an AST Tree next to the corresponding source code.

    Selecting a node in the tree will highlight the corresponing text section in source.

    ``` py
    widget = custom_widgets.AstViewer()
    widget.show()
    ```

    <figure markdown>
      ![Image title](../../images/astviewer.png)
      <figcaption>AstViewer widget</figcaption>
    </figure>
    """

    def __init__(
        self,
        object_name="ast_viewer",
        **kwargs,
    ):
        super().__init__(orientation="horizontal", object_name=object_name, **kwargs)
        self.tree = widgets.TreeView()
        self.ast_textedit = widgets.PlainTextEdit()
        self.tabwidget = widgets.TabWidget()
        self.tabwidget.add_tab(self.tree, "Tree")
        self.tabwidget.add_tab(self.ast_textedit, "Dumped")

        self.textedit = custom_widgets.CodeEditor()
        self.textedit.set_syntaxhighlighter("python")
        self.textedit.textChanged.connect(self._on_text_change)
        self.add(self.tabwidget)
        self.add(self.textedit)

    def _on_text_change(self):
        self.set_ast(self.textedit.get_value())

    def _on_current_change(self, new_index: core.ModelIndex, _):
        node = new_index.data(constants.USER_ROLE)
        if not hasattr(node, "lineno"):
            return
        with self.textedit.selecter.current_cursor() as cursor:
            cursor.select_text(
                (node.lineno - 1, node.col_offset),
                (node.end_lineno - 1, node.end_col_offset),
            )

    def set_ast(self, ast_or_str: str | ast.AST):
        if self.tree.model() is None:
            model = itemmodels.AstModel(ast_or_str, show_root=True, parent=self.tree)
            self.tree.set_model(model)
            self.tree.selectionModel().currentChanged.connect(self._on_current_change)
        else:
            self.tree.model().set_ast(ast_or_str)
        self.tree.show_root(False)
        self.tree.expandAll()
        ast_tree = self.tree.get_model(skip_proxies=True).ast_tree
        text = ast_or_str if isinstance(ast_or_str, str) else ast.unparse(ast_or_str)
        with self.textedit.signals_blocked():
            self.textedit.set_text(text)
        dumped = ast.dump(ast_tree, indent=4)
        self.ast_textedit.set_text(dumped)

    def rename_variable(
        self,
        old: str,
        new: str,
        ignore: list[str] | None = None,
        scope: list[str] | None = None,
    ):
        if scope is None:
            scope = ["main"]
        if ignore is None:
            ignore = []
        tree = self.tree.get_model(skip_proxies=True).ast_tree
        for i in ast.iter_fields(tree):
            if not isinstance(a := getattr(tree, i), list) and (
                a == old and not {*scope} & {*ignore}
            ):
                setattr(tree, i, new)
            n = a if isinstance(a, list) else [a]
            s = [tree.name] if tree.__class__.__name__.endswith("Def") else scope
            for j in n:
                if isinstance(j, ast.AST):
                    self.rename_variable(j, old, new, ignore, s)

    # @classmethod
    # def setup_example(cls):
    #     import pathlib

    #     widget = cls()
    #     code = pathlib.Path(__file__).read_text("utf-8")
    #     widget.set_ast(code)
    #     return widget


if __name__ == "__main__":
    app = widgets.app()
    w = AstViewer.setup_example()
    w.show()
    with app.debug_mode():
        app.exec()



"""Module containing stuff to help buidling docs.

This module builds on mknodes and contains some new nodes to help building documentation,
an MkItemModelTable to easily display Qt ItemModels as Markdown Tables
as well as a model + widget to preview a generated mknodes tree.

"""

from __future__ import annotations

from .qtlinkprovider import QtLinkProvider
from .markdownmodel import MkNodesModel
from .markdownwidget import MarkdownWidget
from .mkitemmodeltable import MkItemModelTable
from .mkprettyqtdiagram import MkPrettyQtDiagram
from .mkpropertytable import MkPropertyTable
from .mkwidgetscreenshot import MkWidgetScreenShot


__all__ = [
    "MarkdownWidget",
    "MkItemModelTable",
    "MkNodesModel",
    "MkPrettyQtDiagram",
    "MkPropertyTable",
    "MkWidgetScreenShot",
    "QtLinkProvider",
]



from __future__ import annotations

from typing import TYPE_CHECKING, Literal

from prettyqt import core, network
from prettyqt.utils import bidict


if TYPE_CHECKING:
    from prettyqt.utils import datatypes


OperationStr = Literal["head", "get", "put", "post", "delete", "custom"]

OPERATION: bidict[OperationStr, network.QNetworkAccessManager.Operation] = bidict(
    head=network.QNetworkAccessManager.Operation.HeadOperation,
    get=network.QNetworkAccessManager.Operation.GetOperation,
    put=network.QNetworkAccessManager.Operation.PutOperation,
    post=network.QNetworkAccessManager.Operation.PostOperation,
    delete=network.QNetworkAccessManager.Operation.DeleteOperation,
    custom=network.QNetworkAccessManager.Operation.CustomOperation,
)


class NetworkAccessManager(core.ObjectMixin, network.QNetworkAccessManager):
    """Allows the application to send network requests and receive replies."""

    # def request(
    #     self,
    #     method,
    #     url,
    #     headers=None,
    #     cookies=None,
    #     files=None,
    #     auth=None,
    #     timeout=None,
    #     allow_redirects=True,
    # ):
    #     req = network.NetworkRequest()
    #     if allow_redirects:
    #         pass

    def get(self, request: datatypes.UrlType | network.QNetworkRequest):
        if isinstance(request, str):
            request = core.Url(request)
        request = network.NetworkRequest(request)
        return super().get(request)

    # def post(self, url, data=None, json=None):
    #     pass

    # def put(self, url, data=None, json=None):
    #     pass

    # def patch(self, url, data=None):
    #     pass

    # def delete(self, url):
    #     pass

    def set_redirect_policy(
        self,
        policy: network.networkrequest.RedirectPolicyStr
        | network.NetworkRequest.RedirectPolicy,
    ):
        """Set redirect policy.

        Args:
            policy: redirect policy
        """
        self.setRedirectPolicy(
            network.networkrequest.REDIRECT_POLICIES.get_enum_value(policy)
        )

    def get_redirect_policy(self) -> network.networkrequest.RedirectPolicyStr:
        """Get the current redirect policy.

        Returns:
            redirect policy
        """
        return network.networkrequest.REDIRECT_POLICIES.inverse[self.redirectPolicy()]



from __future__ import annotations

from prettyqt import core, gui
from prettyqt.utils import datatypes


class PdfWriter(core.ObjectMixin, gui.PagedPaintDeviceMixin, gui.QPdfWriter):
    """Class to generate PDFs that can be used as a paint device."""

    def set_desktop_resolution(self):
        primary = gui.app().primaryScreen()
        dpi = primary.logicalDotsPerInch()
        self.setResolution(int(dpi))

    def set_page_margins(
        self,
        margins: datatypes.MarginsFType,
        unit: gui.pagelayout.UnitStr | gui.PageLayout.Unit | None = None,
    ) -> bool:
        margins = datatypes.to_marginsf(margins)
        if unit is None:
            return self.setPageMargins(margins)
        return self.setPageMargins(margins, gui.pagelayout.UNITS.get_enum_value(unit))

    def get_pdf_version(self) -> gui.pagedpaintdevice.PdfVersionStr:
        return gui.pagedpaintdevice.PDF_VERSION.inverse[self.pdfVersion()]

    def set_pdf_version(
        self,
        version: gui.pagedpaintdevice.PdfVersionStr | gui.PagedPaintDevice.PdfVersion,
    ):
        """Set pdf version.

        Args:
            version: pdf version
        """
        self.setPdfVersion(gui.pagedpaintdevice.PDF_VERSION.get_enum_value(version))


if __name__ == "__main__":
    app = gui.app()
    writer = PdfWriter("")
    writer.set_desktop_resolution()



from __future__ import annotations

import datetime

from prettyqt import core, widgets
from prettyqt.utils import datatypes


class TimeEdit(widgets.DateTimeEditMixin, widgets.QTimeEdit):
    """Widget for editing times based on the QDateTimeEdit widget."""

    value_changed = core.Signal(datetime.datetime)

    def set_range(self, lower: datatypes.TimeType, upper: datatypes.TimeType):
        # self.setToolTip(f"{lower.toString()} <= x <= {upper.toString()}")
        self.setTimeRange(datatypes.to_time(lower), datatypes.to_time(upper))

    def get_value(self) -> datetime.time:
        return self.get_time()

    def set_value(self, value: datatypes.TimeType):
        self.set_time(value)

    def set_time(self, value: datatypes.TimeType):
        super().setTime(datatypes.to_time(value))

    @classmethod
    def setup_example(cls):
        widget = TimeEdit()
        widget.set_value("02:11:50")
        return widget


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    widget = TimeEdit()
    widget.set_time("02:04:10")
    widget.show()
    app.exec()



from __future__ import annotations

import pathlib
from typing import Literal

from prettyqt import quick
from prettyqt.utils import bidict, datatypes


ResizeModeStr = Literal["view_to_root_object", "root_object_to_view"]

RESIZE_MODES: bidict[ResizeModeStr, quick.QQuickView.ResizeMode] = bidict(
    view_to_root_object=quick.QQuickView.ResizeMode.SizeViewToRootObject,
    root_object_to_view=quick.QQuickView.ResizeMode.SizeRootObjectToView,
)

StatusStr = Literal["null", "ready", "loading", "error"]

STATUS: bidict[StatusStr, quick.QQuickView.Status] = bidict(
    null=quick.QQuickView.Status.Null,
    ready=quick.QQuickView.Status.Ready,
    loading=quick.QQuickView.Status.Loading,
    error=quick.QQuickView.Status.Error,
)


class QuickView(quick.quickwindow.QuickWindowMixin, quick.QQuickView):
    """Window for displaying a Qt Quick user interface."""

    def set_source(self, source: datatypes.UrlType | datatypes.PathType):
        self.setSource(datatypes.to_local_url(source))

    def get_source(self) -> pathlib.Path:
        return pathlib.Path(self.source().toLocalFile())

    def get_status(self) -> StatusStr:
        return STATUS.inverse[self.status()]



from __future__ import annotations

from typing import TYPE_CHECKING, Literal, Self

from prettyqt import gui
from prettyqt.utils import bidict, colors, get_repr, serializemixin


if TYPE_CHECKING:
    from prettyqt.utils import datatypes


RoleStr = Literal[
    "window",
    "window_text",
    "base",
    "alternate_base",
    "tool_tip_base",
    "tool_tip_text",
    "placeholder_text",
    "text",
    "button",
    "button_text",
    "bright_text",
    "light",
    "midlight",
    "dark",
    "mid",
    "shadow",
    "highlight",
    "highlighted_text",
    "link",
    "link_visited",
    "none",
]


ROLE: bidict[RoleStr, gui.QPalette.ColorRole] = bidict(
    window=gui.QPalette.ColorRole.Window,  # same as Background
    window_text=gui.QPalette.ColorRole.WindowText,  # same as Foreground
    base=gui.QPalette.ColorRole.Base,
    alternate_base=gui.QPalette.ColorRole.AlternateBase,
    tool_tip_base=gui.QPalette.ColorRole.ToolTipBase,
    tool_tip_text=gui.QPalette.ColorRole.ToolTipText,
    placeholder_text=gui.QPalette.ColorRole.PlaceholderText,
    text=gui.QPalette.ColorRole.Text,
    button=gui.QPalette.ColorRole.Button,
    button_text=gui.QPalette.ColorRole.ButtonText,
    bright_text=gui.QPalette.ColorRole.BrightText,
    light=gui.QPalette.ColorRole.Light,
    midlight=gui.QPalette.ColorRole.Midlight,
    dark=gui.QPalette.ColorRole.Dark,
    mid=gui.QPalette.ColorRole.Mid,
    shadow=gui.QPalette.ColorRole.Shadow,
    highlight=gui.QPalette.ColorRole.Highlight,
    highlighted_text=gui.QPalette.ColorRole.HighlightedText,
    link=gui.QPalette.ColorRole.Link,
    link_visited=gui.QPalette.ColorRole.LinkVisited,
    none=gui.QPalette.ColorRole.NoRole,
)

GroupStr = Literal["disabled", "active", "inactive"]

GROUP: bidict[GroupStr, gui.QPalette.ColorGroup] = bidict(
    disabled=gui.QPalette.ColorGroup.Disabled,
    active=gui.QPalette.ColorGroup.Active,  # normal
    inactive=gui.QPalette.ColorGroup.Inactive,
)


class Palette(serializemixin.SerializeMixin, gui.QPalette):
    """Contains color groups for each widget state."""

    def __getitem__(self, index: RoleStr | gui.QPalette.ColorRole) -> gui.Color:
        return self.get_color(index)

    def __setitem__(
        self, index: RoleStr | gui.QPalette.ColorRole, value: datatypes.ColorType
    ):
        self.set_color(index, value)

    def __repr__(self):
        return get_repr(self, self.get_color("button"), self.get_color("window"))

    def iter_colors(
        self, mode: Literal["auto", "auto_inverted", "dark", "light"] = "auto"
    ):
        if mode == "auto":
            mode = "dark" if self.is_dark() else "light"
        elif mode == "auto_inverted":
            mode = "light" if self.is_dark() else "dark"
        for color in gui.Color.colorNames():
            c = gui.Color(color)
            if (c.is_dark() and mode == "dark") or (not c.is_dark() and mode == "light"):
                yield c.as_qt()

    def highlight_inactive(self, enable: bool = True):
        if enable:
            color = self.color(self.ColorGroup.Active, self.ColorRole.Highlight)
        else:
            pal = gui.Palette()
            color = pal.color(self.ColorGroup.Inactive, self.ColorRole.Highlight)
        self.setColor(self.ColorGroup.Inactive, self.ColorRole.Highlight, color)

    def set_color(
        self,
        role: RoleStr | gui.QPalette.ColorRole,
        color: datatypes.ColorType,
        group: GroupStr | gui.QPalette.ColorGroup = "active",
    ):
        color = colors.get_color(color)
        self.setColor(GROUP.get_enum_value(group), ROLE.get_enum_value(role), color)

    def get_colors(
        self, group: GroupStr | gui.QPalette.ColorGroup = "active"
    ) -> dict[str, gui.Color]:
        return {k: self.get_color(k, group) for k in ROLE}

    def get_color(
        self,
        role: RoleStr | gui.QPalette.ColorRole,
        group: GroupStr | gui.QPalette.ColorGroup = "active",
    ) -> gui.Color:
        return gui.Color(
            self.color(GROUP.get_enum_value(group), ROLE.get_enum_value(role))
        )

    def set_brush(
        self,
        role: RoleStr | gui.QPalette.ColorRole,
        brush: datatypes.ColorAndBrushType,
        group: GroupStr | gui.QPalette.ColorGroup = "active",
    ):
        if not isinstance(brush, gui.QBrush):
            brush = gui.Brush(colors.get_color(brush))
        self.setBrush(GROUP.get_enum_value(group), ROLE.get_enum_value(role), brush)

    def get_brushes(
        self, group: GroupStr | gui.QPalette.ColorGroup = "active"
    ) -> dict[str, gui.Brush]:
        return {k: self.get_brush(k, group) for k in ROLE}

    def get_brush(
        self,
        role: RoleStr | gui.QPalette.ColorRole,
        group: GroupStr | gui.QPalette.ColorGroup = "active",
    ) -> gui.Brush:
        return gui.Brush(
            self.brush(GROUP.get_enum_value(group), ROLE.get_enum_value(role))
        )

    def set_color_group(
        self, group: GroupStr | gui.QPalette.ColorGroup, **kwargs: gui.QBrush
    ):
        """Set the color group.

        Args:
            group: color group to use
            kwargs: keyword arguments passed to setColorGroup
        """
        self.setColorGroup(GROUP.get_enum_value(group), **kwargs)

    def get_color_group(self) -> GroupStr:
        """Return color group.

        Returns:
            color group
        """
        return GROUP.inverse[self.colorGroup()]

    def inverted(self) -> Self:
        pal = type(self)()
        for group in GROUP:
            for role in ROLE:
                color = self.get_color(role, group)
                pal.set_color(role, color.inverted(), group)
        return pal

    @classmethod
    def create_dark_palette(cls) -> Self:
        pal = cls()
        pal.set_color("window", gui.Color(53, 53, 53))
        pal.set_color("window_text", "white")
        pal.set_color("window_text", "grey", group="disabled")
        pal.set_color("base", gui.Color(25, 25, 25))
        pal.set_color("alternate_base", gui.Color(53, 53, 53))
        pal.set_color("tool_tip_base", "slategrey")
        pal.set_color("tool_tip_base", "slategrey", group="inactive")
        pal.set_color("tool_tip_text", "white")
        pal.set_color("tool_tip_text", "white", group="inactive")
        pal.set_color("text", "white")
        pal.set_color("text", "grey", group="disabled")
        pal.set_color("button", gui.Color(53, 53, 53))
        pal.set_color("button_text", "white")
        pal.set_color("button_text", "grey", group="disabled")
        pal.set_color("bright_text", "red")
        pal.set_color("link", "dodgerblue")
        pal.set_color("highlight", "dodgerblue")
        pal.set_color("highlight", gui.Color(80, 80, 80), group="disabled")
        pal.set_color("highlighted_text", "black")
        pal.set_color("highlighted_text", "grey", group="disabled")
        return pal

    def is_dark(self) -> bool:
        col = self.color(self.ColorRole.Window)
        return max(col.getRgb()[:3]) < 115  # noqa: PLR2004


if __name__ == "__main__":
    pal = Palette()
    print(list(pal.yield_dark_colors()))



from __future__ import annotations

from typing import Any

from prettyqt import widgets


class MappedCheckBox(widgets.CheckBox):
    def __init__(
        self,
        *args,
        true_value=True,
        false_value=False,
        object_name: str = "mapped_checkbox",
        **kwargs,
    ):
        import bidict

        super().__init__(*args, object_name=object_name, **kwargs)
        dct = {True: true_value, False: false_value}
        self.map = bidict.bidict[bool, Any](dct)

    def get_value(self):
        return self.map[self.isChecked()]

    def set_value(self, value):
        val = self.map.inverse[value]
        super().set_value(val)


if __name__ == "__main__":
    app = widgets.app()
    widget = MappedCheckBox("Test")
    widget.show()
    app.exec()
    print(widget.get_value())



from __future__ import annotations

from typing import Literal

from prettyqt import core, location
from prettyqt.utils import bidict


QGeoServiceProvider = location.QGeoServiceProvider

ErrorStr = Literal[
    "none",
    "not_supported",
    "unknown_parameter",
    "missing_required_parameter",
    "connection",
    "failed_to_load",
]

ERROR: bidict[ErrorStr, QGeoServiceProvider.Error] = bidict(
    none=QGeoServiceProvider.Error.NoError,
    not_supported=QGeoServiceProvider.Error.NotSupportedError,
    unknown_parameter=QGeoServiceProvider.Error.UnknownParameterError,
    missing_required_parameter=QGeoServiceProvider.Error.MissingRequiredParameterError,
    connection=QGeoServiceProvider.Error.ConnectionError,
    failed_to_load=QGeoServiceProvider.Error.LoaderError,
)

GeocodingFeatureStr = Literal[
    "none",
    "online",
    "offline",
    "reverse",
    "localized",
]

GEOCODING_FEATURES: bidict[GeocodingFeatureStr, QGeoServiceProvider.GeocodingFeature] = (
    bidict(
        none=QGeoServiceProvider.GeocodingFeature.NoGeocodingFeatures,
        online=QGeoServiceProvider.GeocodingFeature.OnlineGeocodingFeature,
        offline=QGeoServiceProvider.GeocodingFeature.OfflineGeocodingFeature,
        reverse=QGeoServiceProvider.GeocodingFeature.ReverseGeocodingFeature,
        localized=QGeoServiceProvider.GeocodingFeature.LocalizedGeocodingFeature,
        # any=QGeoServiceProvider.GeocodingFeature.AnyGeocodingFeatures,
    )
)

MappingFeatureStr = Literal[
    "none",
    "online",
    "offline",
    "localized",
]

MAPPING_FEATURES: bidict[MappingFeatureStr, QGeoServiceProvider.MappingFeature] = bidict(
    none=QGeoServiceProvider.MappingFeature.NoMappingFeatures,
    online=QGeoServiceProvider.MappingFeature.OnlineMappingFeature,
    offline=QGeoServiceProvider.MappingFeature.OfflineMappingFeature,
    localized=QGeoServiceProvider.MappingFeature.LocalizedMappingFeature,
    # any=QGeoServiceProvider.MappingFeature.AnyMappingFeatures,
)

NavigationFeatureStr = Literal[
    "none",
    "online",
    "offline",
]

NAVIGATION_FEATURES: bidict[
    NavigationFeatureStr, QGeoServiceProvider.NavigationFeature
] = bidict(
    none=QGeoServiceProvider.NavigationFeature.NoNavigationFeatures,
    online=QGeoServiceProvider.NavigationFeature.OnlineNavigationFeature,
    offline=QGeoServiceProvider.NavigationFeature.OfflineNavigationFeature,
    # any=QGeoServiceProvider.NavigationFeature.AnyNavigationFeatures,
)

PlaceFeatureStr = Literal[
    "none",
    "online_places",
    "offline_places",
    "save_place",
    "remove_place",
    "save_category",
    "remove_category",
    "place_recommendations",
    "search_suggestions",
    "localized_places",
    "notifications",
    "place_matching",
]

PLACES_FEATURES: bidict[PlaceFeatureStr, QGeoServiceProvider.PlacesFeature] = bidict(
    none=QGeoServiceProvider.PlacesFeature.NoPlacesFeatures,
    online_places=QGeoServiceProvider.PlacesFeature.OnlinePlacesFeature,
    offline_places=QGeoServiceProvider.PlacesFeature.OfflinePlacesFeature,
    save_place=QGeoServiceProvider.PlacesFeature.SavePlaceFeature,
    remove_place=QGeoServiceProvider.PlacesFeature.RemovePlaceFeature,
    save_category=QGeoServiceProvider.PlacesFeature.SaveCategoryFeature,
    remove_category=QGeoServiceProvider.PlacesFeature.RemoveCategoryFeature,
    place_recommendations=QGeoServiceProvider.PlacesFeature.PlaceRecommendationsFeature,
    search_suggestions=QGeoServiceProvider.PlacesFeature.SearchSuggestionsFeature,
    localized_places=QGeoServiceProvider.PlacesFeature.LocalizedPlacesFeature,
    notifications=QGeoServiceProvider.PlacesFeature.NotificationsFeature,
    place_matching=QGeoServiceProvider.PlacesFeature.PlaceMatchingFeature,
    # any=QGeoServiceProvider.PlacesFeature.AnyPlacesFeatures,
)

RoutingFeatureStr = Literal[
    "none",
    "online",
    "offline",
    "localized",
    "route_updates",
    "alternative_routes",
    "exclude_areas",
]

ROUTING_FEATURES: bidict[RoutingFeatureStr, QGeoServiceProvider.RoutingFeature] = bidict(
    none=QGeoServiceProvider.RoutingFeature.NoRoutingFeatures,
    online=QGeoServiceProvider.RoutingFeature.OnlineRoutingFeature,
    offline=QGeoServiceProvider.RoutingFeature.OfflineRoutingFeature,
    localized=QGeoServiceProvider.RoutingFeature.LocalizedRoutingFeature,
    route_updates=QGeoServiceProvider.RoutingFeature.RouteUpdatesFeature,
    alternative_routes=QGeoServiceProvider.RoutingFeature.AlternativeRoutesFeature,
    exclude_areas=QGeoServiceProvider.RoutingFeature.ExcludeAreasRoutingFeature,
    # any=location.QGeoServiceProvider.RoutingFeature.AnyRoutingFeatures,
)


class GeoServiceProvider(core.ObjectMixin, location.QGeoServiceProvider):
    """Aggregates access to services which provide geographical information."""

    def get_error(self) -> ErrorStr:
        return ERROR.inverse[self.error()]

    def get_geocoding_error(self) -> ErrorStr:
        return ERROR.inverse[self.geocodingError()]

    def get_geocoding_features(self) -> list[GeocodingFeatureStr]:
        return GEOCODING_FEATURES.get_list(self.geocodingFeatures())

    def get_mapping_error(self) -> ErrorStr:
        return ERROR.inverse[self.mappingError()]

    def get_mapping_features(self) -> list[MappingFeatureStr]:
        return MAPPING_FEATURES.get_list(self.mappingFeatures())

    def get_navigation_error(self) -> ErrorStr:
        return ERROR.inverse[self.navigationError()]

    def get_navigation_features(self) -> list[NavigationFeatureStr]:
        return NAVIGATION_FEATURES.get_list(self.navigationFeatures())

    def get_places_error(self) -> ErrorStr:
        return ERROR.inverse[self.placesError()]

    def get_places_features(self) -> list[PlaceFeatureStr]:
        return PLACES_FEATURES.get_list(self.placesFeatures())

    def get_routing_error(self) -> ErrorStr:
        return ERROR.inverse[self.routingError()]

    def get_routing_features(self) -> list[RoutingFeatureStr]:
        return ROUTING_FEATURES.get_list(self.routingFeatures())

    def get_geocoding_manager(self) -> location.GeoCodingManager:
        return location.GeoCodingManager(self.geocodingManager())

    def get_routing_manager(self) -> location.GeoRoutingManager:
        return location.GeoRoutingManager(self.routingManager())

    def get_place_manager(self) -> location.PlaceManager:
        return location.PlaceManager(self.placeManager())



from __future__ import annotations

from prettyqt.qt import QtDesigner


class PyDesignerCustomWidgetCollection(QtDesigner.QPyDesignerCustomWidgetCollection):
    pass


if __name__ == "__main__":
    fact = PyDesignerCustomWidgetCollection()



from __future__ import annotations

from typing import Literal

from prettyqt import core, gui, widgets
from prettyqt.utils import bidict


FontFilterStr = Literal["all", "scalable", "non_scalable", "monospaced", "proportional"]

FONT_FILTERS: bidict[FontFilterStr, widgets.QFontComboBox.FontFilter] = bidict(
    all=widgets.QFontComboBox.FontFilter.AllFonts,
    scalable=widgets.QFontComboBox.FontFilter.ScalableFonts,
    non_scalable=widgets.QFontComboBox.FontFilter.NonScalableFonts,
    monospaced=widgets.QFontComboBox.FontFilter.MonospacedFonts,
    proportional=widgets.QFontComboBox.FontFilter.ProportionalFonts,
)


class FontComboBox(widgets.ComboBoxMixin, widgets.QFontComboBox):
    """Combobox that lets the user select a font family."""

    value_changed = core.Signal(gui.QFont)

    def _get_map(self):
        maps = super()._get_map()
        maps |= {
            "fontFilters": FONT_FILTERS,
            "writingSystem": gui.fontdatabase.WRITING_SYSTEM,
        }
        return maps

    def set_font_filters(self, *filters: FontFilterStr):
        """Set font filters.

        Args:
            filters: font filters to use
        """
        if not filters:
            filters = ("all",)
        flags = FONT_FILTERS.merge_flags(filters)
        self.setFontFilters(flags)

    def get_font_filters(self) -> list[FontFilterStr]:
        """Return list of font filters.

        Returns:
            font filter list
        """
        return FONT_FILTERS.get_list(self.fontFilters())

    def set_value(self, value: gui.QFont):
        self.setCurrentFont(value)

    def get_value(self) -> gui.QFont:
        return gui.QFont(self.get_current_font())

    def get_current_font(self) -> gui.Font:
        return gui.Font(self.currentFont())

    # without this, the user property would be currentText, which is not what we want.
    # PyQt6 doesnt like gui.Font here.
    value = core.Property(
        gui.QFont,
        get_value,
        set_value,
        user=True,
        doc="Currently chosen font",
    )


if __name__ == "__main__":
    app = widgets.app()
    widget = FontComboBox()
    widget.set_value(gui.Font("Script"))
    widget.show()
    app.exec()



from __future__ import annotations

from prettyqt import core, gui, network, widgets


class BaseScrapeModel(gui.StandardItemModel):
    finished = core.Signal()
    error = core.Signal(str)
    search_url: str

    def __init__(self, parent: core.QObject | None = None):
        super().__init__(parent)
        self._manager = network.NetworkAccessManager(self)
        self._reply = None

    @core.Slot(str)
    def search(self, text: str):
        self.clear()
        if self._reply is not None:
            self._reply.abort()
        if text:
            r = self.search_url.format(text=text)
            self._reply = self._manager.get(r)
            self._reply.finished.connect(self.on_finished)
            loop = core.EventLoop()
            self.finished.connect(loop.quit)
            loop.exec()

    @core.Slot()
    def on_finished(self):
        assert self._reply
        if self._reply.error() == network.QNetworkReply.NetworkError.NoError:
            data = self._reply.readAll().data()
            assert isinstance(data, bytes)
            response = data.decode()
            for s in self.process_reply(response):
                self.appendRow(gui.StandardItem(s))
        self.finished.emit()
        # self._reply.deleteLater()

    def process_reply(self, reply: str) -> list[str]:
        return NotImplemented


class GoogleSearchModel(BaseScrapeModel):
    search_url = "https://google.com/complete/search?output=toolbar&q={text}"

    def process_reply(self, reply: str) -> list[str]:
        return [
            xml.attributes().value("data")
            for xml in core.XmlStreamReader(reply)
            if (
                xml.tokenType() == core.XmlStreamReader.TokenType.StartElement
                and xml.name() == "suggestion"
            )
        ]


class GoogleCompleter(widgets.Completer):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.set_completion_mode("unfiltered_popup")
        self.set_case_sensitive(False)
        self._model = GoogleSearchModel()
        self.setModel(self._model)
        self.setCompletionPrefix("")

    def splitPath(self, path: str):
        self._model.search(path)
        return super().splitPath(path)


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    wnd = widgets.Widget()
    wnd.set_layout("horizontal")
    widget = widgets.LineEdit()
    wnd.box.add(widget)
    completer = GoogleCompleter(widget)
    widget.set_completer(completer)
    wnd.show()
    app.exec()



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, ClassVar

import pydantic

from prettyqt import constants, itemmodels
from prettyqt.qt import QtGui


if TYPE_CHECKING:
    from prettyqt import core


logger = logging.getLogger(__name__)


class PydanticFieldsModel(itemmodels.BaseFieldsModel):
    """Table model to display the fields and their metadata of a pydantic model.

    More information about pydantic can be found [here][https://www.pydantic.dev/].

    Frozen BaseModels / frozen fields read-only.

    !!! note
        This ItemModel only supports Pydantic v2 BaseModels.
    """

    SUPPORTS = pydantic.BaseModel
    HEADER: ClassVar = [
        "Value",
        "Type",
        "Default",
        "Frozen",
        "Alias",
        "in repr",
    ]

    def __init__(self, instance: pydantic.BaseModel, **kwargs):
        super().__init__(instance, **kwargs)

    @classmethod
    def supports(cls, instance) -> bool:
        return isinstance(instance, pydantic.BaseModel)

    def get_fields(self, instance: pydantic.BaseModel):
        return instance.model_fields

    def data(  # noqa: PLR0911
        self,
        index: core.ModelIndex,
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
    ):
        if not index.isValid():
            return None
        field_name = self._field_names[index.row()]
        field = self._fields[field_name]
        value = getattr(self._instance, field_name)
        match role, index.column():
            case constants.DISPLAY_ROLE, 0:
                return repr(value)
            case constants.EDIT_ROLE, 0:
                return value
            case constants.FONT_ROLE, 0:
                font = QtGui.QFont()
                font.setBold(True)
                return font
            case constants.DISPLAY_ROLE, 1:
                return repr(field.annotation)
            case constants.FONT_ROLE, 1:
                font = QtGui.QFont()
                font.setItalic(True)
                return font
            case constants.DISPLAY_ROLE, 2:
                return field.default
            case constants.CHECKSTATE_ROLE, 3:
                return self.to_checkstate(bool(field.frozen))
            case constants.DISPLAY_ROLE, 4:
                return field.alias
            case constants.CHECKSTATE_ROLE, 5:
                return self.to_checkstate(field.repr)
            case constants.USER_ROLE, _:
                return value

    def _is_writable(self, field_name: str):
        model_frozen = self._instance.model_config.get("frozen")
        field_frozen = self._instance.model_fields[field_name].frozen
        frozen = model_frozen or field_frozen
        return not frozen


if __name__ == "__main__":
    from prettyqt import widgets

    class BaseSetting(pydantic.BaseModel):
        name: str = pydantic.Field("test", frozen=True)
        label: str
        description: str
        requires_restart: bool = False
        options: list[str] | None = None
        minimum: float = 0.0
        maximum: float = 0.0
        pattern: str = ""

    class SelectionSetting(BaseSetting):
        options: list[str] | None = None
        minimum: float = 0.0
        maximum: float = 0.0
        pattern: str = ""
        requires_restart: bool = False

    app_style = BaseSetting(
        name="app_style",
        label="App style",
        description="Some longer text",
        options=[],
        requires_restart=True,
    )
    app = widgets.app()
    view = widgets.TableView()
    view.set_icon("mdi.folder")
    with app.debug_mode():
        model = PydanticFieldsModel(app_style, parent=view)
        view.set_model(model)
        view.set_selection_behavior("rows")
        view.setEditTriggers(view.EditTrigger.AllEditTriggers)
        view.set_delegate("editor", column=0)
        view.show()
        view.resize(500, 300)
        app.exec()



from __future__ import annotations

from prettyqt import custom_widgets, validators, widgets
from prettyqt.utils import datatypes


class RegexLineEdit(custom_widgets.SingleLineTextEdit):
    def __init__(self, *args, object_name: str = "regex_lineedit", **kwargs):
        super().__init__(*args, object_name=object_name, **kwargs)
        self.set_syntaxhighlighter("regex")
        self.set_margin(0)
        val = validators.RegexPatternValidator()
        self.set_validator(val)

    def set_value(self, value: datatypes.PatternAndStringType | None):
        match value:
            case None:
                self.set_text("")
            case _:
                self.set_text(datatypes.to_py_pattern(value).pattern)


if __name__ == "__main__":
    app = widgets.app()
    widget = RegexLineEdit()
    widget.show()
    widget.value_changed.connect(print)
    app.exec()



from __future__ import annotations

from prettyqt import core, network


class TcpServer(core.ObjectMixin, network.QTcpServer):
    """TCP-based server."""

    def listen(  # type: ignore
        self, address: str | network.QHostAddress, port: int = 0
    ) -> bool:
        if isinstance(address, str):
            address = network.HostAddress(address)
        return super().listen(address, port)

    def get_server_address(self) -> network.HostAddress:
        return network.HostAddress(self.serverAddress())

    def get_proxy(self) -> network.NetworkProxy:
        return network.NetworkProxy(self.proxy())

    def get_server_error(self) -> network.abstractsocket.SocketErrorStr:
        return network.abstractsocket.SOCKET_ERROR.inverse[self.serverError()]



from __future__ import annotations

import contextlib
from typing import Literal

from prettyqt import core
from prettyqt.utils import bidict


mod = core.QIODeviceBase  # type: ignore

OpenModeStr = Literal[
    "not_open",
    "read_only",
    "write_only",
    "read_write",
    "append",
    "truncate",
    "text",
    "unbuffered",
    "new_only",
    "existing_only",
]

OPEN_MODES: Literal[OpenModeStr, mod.OpenModeFlag] = bidict(
    not_open=mod.OpenModeFlag.NotOpen,
    read_only=mod.OpenModeFlag.ReadOnly,
    write_only=mod.OpenModeFlag.WriteOnly,
    read_write=mod.OpenModeFlag.ReadWrite,
    append=mod.OpenModeFlag.Append,
    truncate=mod.OpenModeFlag.Truncate,
    text=mod.OpenModeFlag.Text,
    unbuffered=mod.OpenModeFlag.Unbuffered,
    new_only=mod.OpenModeFlag.NewOnly,
    existing_only=mod.OpenModeFlag.ExistingOnly,
)


class IODeviceMixin(core.ObjectMixin):
    def __len__(self):
        return self.size()

    @contextlib.contextmanager
    def open_file(self, mode: OpenModeStr | mod.OpenModeFlag):
        self.open(OPEN_MODES.get_enum_value(mode))
        yield self
        self.close()

    def get_open_mode(self) -> OpenModeStr:
        return OPEN_MODES.inverse[self.openMode()]


class IODevice(IODeviceMixin, core.QIODevice):
    """The base interface class of all I/O devices in Qt."""



from __future__ import annotations

import logging
from typing import Any, Literal

import numpy as np

from prettyqt import constants, core


logger = logging.getLogger(__name__)


RankMethod = Literal["average", "min", "max", "dense", "ordinal"]


class RankProxyModel(core.AbstractProxyModel):
    """Proxy model to compute numerical ranks for values.

    === "Without proxy"

        ```py
        app = widgets.app()
        data = {
            "Name": ["John", "Alice", "Bob", "Carol", "Dave"],
            "Score": [85, 92, 78, 92, 88],
            "Time": [120, 95, 150, 95, 110],
        }
        model = gui.StandardItemModel.from_dict(data)
        table = widgets.TableView()
        table.set_model(model)
        table.show()
        ```

    === "With proxy"

        ```py
        app = widgets.app()
        data = {
            "Name": ["John", "Alice", "Bob", "Carol", "Dave"],
            "Score": [85, 92, 78, 92, 88],
            "Time": [120, 95, 150, 95, 110],
        }
        model = gui.StandardItemModel.from_dict(data)
        table = widgets.TableView()
        table.set_model(model)
        table.proxifier.rank(
            column=1,  # Score column
            method="dense",  # ranking method
            ascending=False  # higher is better
        )
        table.show()
        ```
    """

    ID = "rank"
    ICON = "mdi6.order-numeric-ascending"

    def __init__(
        self,
        column: int,
        method: RankMethod = "average",
        ascending: bool = True,
        na_option: Literal["keep", "top", "bottom"] = "keep",
        **kwargs,
    ):
        """Initialize RankProxyModel.

        Args:
            column: Column to rank
            method: How to assign rank to equal values
                'average': average rank of the group
                'min': lowest rank in the group
                'max': highest rank in the group
                'dense': like 'min', but rank always increases by 1
                'ordinal': all values get unique rank
            ascending: Whether to rank in ascending order
            na_option: How to handle NA values
                'keep': keep NA values as NA
                'top': assign lowest rank to NA
                'bottom': assign highest rank to NA
            kwargs: Additional keyword arguments
        """
        super().__init__(**kwargs)
        self._column = column
        self._method = method
        self._ascending = ascending
        self._na_option = na_option
        # Cache for ranks since calculation is fast but not instant
        self._ranks: dict[int, float | None] = {}

    def _calculate_ranks(self) -> None:
        """Calculate ranks for current data."""
        source = self.sourceModel()
        if not source:
            return

        # Get values and their indices
        values = []
        indices = []
        na_indices = []

        for row in range(source.rowCount()):
            idx = source.index(row, self._column)
            val = source.data(idx, constants.EDIT_ROLE)
            try:
                val = float(val)
                values.append(val)
                indices.append(row)
            except (ValueError, TypeError):
                na_indices.append(row)

        if not values:
            return

        # Convert to numpy array for efficient ranking
        arr = np.array(values)
        if not self._ascending:
            arr = -arr

        # Calculate ranks
        if self._method == "ordinal":
            ranks = np.argsort(arr).argsort() + 1
        else:
            ties = {
                "average": "average",
                "min": "min",
                "max": "max",
                "dense": "dense",
            }[self._method]
            ranks = np.array(list(range(1, len(arr) + 1)), dtype=float)

            # Handle ties
            unique_values = np.unique(arr)
            for value in unique_values:
                mask = arr == value
                if ties == "average":
                    rank = ranks[mask].mean()
                elif ties == "min":
                    rank = ranks[mask].min()
                elif ties == "max":
                    rank = ranks[mask].max()
                elif ties == "dense":
                    rank = np.where(arr < value)[0].size + 1
                ranks[mask] = rank

        # Store ranks in cache
        self._ranks.clear()
        for idx, rank in zip(indices, ranks):
            self._ranks[idx] = rank

        # Handle NA values
        if na_indices:
            if self._na_option == "top":
                na_rank = 0
            elif self._na_option == "bottom":
                na_rank = len(values) + 1
            else:  # "keep"
                na_rank = None

            for idx in na_indices:
                self._ranks[idx] = na_rank

    def columnCount(self, parent: core.ModelIndex | None = None) -> int:
        """Return number of columns (same as source)."""
        return self.sourceModel().columnCount() if self.sourceModel() else 0

    def rowCount(self, parent: core.ModelIndex | None = None) -> int:
        """Return number of rows (same as source)."""
        return self.sourceModel().rowCount() if self.sourceModel() else 0

    def data(
        self,
        index: core.ModelIndex,
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
    ) -> Any:
        """Return data for given index and role."""
        if not index.isValid():
            return None

        # Only modify the specified column
        if index.column() != self._column:
            return self.sourceModel().data(index, role)

        # Calculate ranks if needed
        if not self._ranks:
            self._calculate_ranks()

        rank = self._ranks.get(index.row())

        if role == constants.DISPLAY_ROLE:
            return str(int(rank)) if rank is not None else ""
        if role == constants.EDIT_ROLE:
            return rank

        return self.sourceModel().data(index, role)

    def headerData(
        self,
        section: int,
        orientation: constants.Orientation,
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
    ) -> Any:
        """Return header data (same as source)."""
        return self.sourceModel().headerData(section, orientation, role)

    def setSourceModel(self, model: core.QAbstractItemModel) -> None:
        """Set source model and clear cache."""
        super().setSourceModel(model)
        self._ranks.clear()

    def sourceDataChanged(
        self,
        top_left: core.ModelIndex,
        bottom_right: core.ModelIndex,
        roles: list[constants.ItemDataRole],
    ) -> None:
        """Handle source data changes."""
        # Only recalculate if the ranked column was changed
        if top_left.column() <= self._column <= bottom_right.column():
            self._ranks.clear()
        super().sourceDataChanged(top_left, bottom_right, roles)

    def mapToSource(self, proxy_index: core.ModelIndex) -> core.ModelIndex:
        """Map proxy index to source index (direct mapping)."""
        if not proxy_index.isValid():
            return core.ModelIndex()
        return self.sourceModel().index(proxy_index.row(), proxy_index.column())

    def mapFromSource(self, source_index: core.ModelIndex) -> core.ModelIndex:
        """Map source index to proxy index (direct mapping)."""
        if not source_index.isValid():
            return core.ModelIndex()
        return self.index(source_index.row(), source_index.column())

    def get_column(self) -> int:
        """Get column to rank."""
        return self._column

    def set_column(self, column: int):
        """Set column to rank."""
        with self.reset_model():
            self._column = column
            self._ranks.clear()

    def get_method(self) -> RankMethod:
        """Get ranking method."""
        return self._method

    def set_method(self, method: RankMethod):
        """Set ranking method."""
        with self.reset_model():
            self._method = method
            self._ranks.clear()

    def get_ascending(self) -> bool:
        """Get sort order."""
        return self._ascending

    def set_ascending(self, ascending: bool):
        """Set sort order."""
        with self.reset_model():
            self._ascending = ascending
            self._ranks.clear()

    column = core.Property(
        int,
        get_column,
        set_column,
        doc="Column to rank",
    )
    method = core.Property(
        str,
        get_method,
        set_method,
        doc="Ranking method",
    )
    ascending = core.Property(
        bool,
        get_ascending,
        set_ascending,
        doc="Sort order",
    )


if __name__ == "__main__":
    from prettyqt import gui, widgets

    app = widgets.app(style="Windows")
    data = {
        "Name": ["John", "Alice", "Bob", "Carol", "Dave"],
        "Score": [85, 92, 78, 92, 88],
        "Time": [120, 95, 150, 95, 110],
    }
    model = gui.StandardItemModel.from_dict(data)
    table = widgets.TableView()
    table.set_model(model)
    table.setWindowTitle("RankProxyModel example")
    table.set_icon("mdi6.order-numeric-ascending")
    table.show()
    table.resize(600, 200)
    table.h_header.resize_sections("stretch")

    # Rank scores (higher is better)
    table.proxifier.rank(
        column=1,  # Score column
        method="dense",  # ranking method
        ascending=False,  # higher is better
    )

    with app.debug_mode():
        app.exec()



"""Classes for displaying the contents of SVG files."""

from prettyqt.qt.QtSvg import *  # noqa: F403

from .svggenerator import SvgGenerator
from .svgrenderer import SvgRenderer
from prettyqt.qt import QtSvg

QT_MODULE = QtSvg

__all__ = ["SvgGenerator", "SvgRenderer"]



from __future__ import annotations

from prettyqt import core, network


class SingleApplicationMixin:
    message_received = core.Signal(str)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # generate a (hopefully) unique name
        self.app_id = type(self).__name__ + self.applicationName()
        self._activate_on_message = True

        # Is there another instance running?
        out_socket = network.LocalSocket()
        out_socket.connect_to_server(self.app_id)
        self._is_running = out_socket.waitForConnected()
        self._in_socket = None
        self._in_stream = None
        if not self._is_running:
            self._out_socket = None
            self._out_stream = None
            self._server = network.LocalServer()
            self._server.listen(self.app_id)
            self._server.newConnection.connect(self._on_new_connection)
        else:
            self._out_socket = out_socket
            self._out_stream = core.TextStream(self._out_socket)
            # self._out_stream.set_codec("UTF-8")

    def is_running(self) -> bool:
        return self._is_running

    def _activate_window(self):
        # bit dumb inheritance check to avoid importing QtWidgets so that this mixin
        # can also get used with QGuiApplication and QCoreApplication
        if hasattr(self, "get_mainwindow"):
            window = self.get_mainwindow()
            if window is None:
                return
            window.raise_to_top()

    def _send_message(self, msg: str) -> bool:
        if self._out_stream is None or self._out_socket is None:
            return False
        self._out_stream << msg << "\n"
        self._out_stream.flush()
        return self._out_socket.waitForBytesWritten()

    def _on_new_connection(self):
        if self._in_socket:
            self._in_socket.readyRead.disconnect(self._on_ready_read)
        self._in_socket = self._server.nextPendingConnection()  # type: ignore
        if self._in_socket is None:
            return
        self._in_stream = core.TextStream(self._in_socket)
        # self._in_stream.set_codec("UTF-8")
        self._in_socket.readyRead.connect(self._on_ready_read)
        if self._activate_on_message:
            self._activate_window()

    def _on_ready_read(self):
        if self._in_stream is None:
            raise RuntimeError
        for msg in self._in_stream.read_lines():
            self.message_received.emit(msg)


if __name__ == "__main__":
    App = type("SingleApp", (SingleApplicationMixin, core.CoreApplication), {})
    app1 = App()
    # app1.exec()

# Alternative:
# logger = logging.getLogger(__name__)


# class SingleApplicationMixin:
#     message_received = core.Signal(str)

#     def __init__(self, *args, **kwargs):
#         super().__init__(*args, **kwargs)
#         # generate a (hopefully) unique name
#         self.key = type(self).__name__ + self.applicationName()
#         self.timeout = 1000
#         self.server = network.LocalServer(self)

#         # cleanup (only needed for unix)
#         core.SharedMemory(self.key).attach()
#         self.memory = core.SharedMemory(self)
#         self.memory.setKey(self.key)

#         if self.memory.attach():
#             self.is_running = True
#             self.send_message(sys.argv[1] if len(sys.argv) > 1 else "show")
#             logger.info("Another instance is already running.")
#             sys.exit(1)

#         self.is_running = False
#         if not self.memory.create(1):
#             logger.error(self.memory.errorString())
#             raise RuntimeError(self.memory.errorString())

#         self.server.newConnection.connect(self._on_new_connection)
#         self.server.listen(self.key)

#     def _on_new_connection(self):
#         socket = self.server.nextPendingConnection()
#         if socket.waitForReadyRead(self.timeout):
#             data = socket.readAll().data().decode()
#             self.message_received.emit(data)
#             socket.disconnectFromServer()

#     def send_message(self, message: str):
#         if not self.is_running:
#             return

#         # connect to another application
#         socket = network.LocalSocket(self)
#         socket.connectToServer(self.key, socket.OpenModeFlag.WriteOnly)
#         if not socket.waitForConnected(self.timeout):
#             logger.error(socket.errorString())
#             return

#         # send message
#         socket.write(message.encode())
#         if not socket.waitForBytesWritten(self.timeout):
#             logger.error(socket.errorString())
#             return

#         socket.disconnectFromServer()


# if __name__ == "__main__":
#     App = type("SingleApp", (SingleApplicationMixin, core.CoreApplication), {})
#     app1 = App(sys.argv)
#     app1.exec()



from __future__ import annotations

from typing import Literal

from prettyqt import constants, core, widgets
from prettyqt.utils import bidict


SubmitPolicyStr = Literal["auto", "manual"]

SUBMIT_POLICY: bidict[SubmitPolicyStr, widgets.QDataWidgetMapper.SubmitPolicy] = bidict(
    auto=widgets.QDataWidgetMapper.SubmitPolicy.AutoSubmit,
    manual=widgets.QDataWidgetMapper.SubmitPolicy.ManualSubmit,
)


class DataWidgetMapper(core.ObjectMixin, widgets.QDataWidgetMapper):
    """Mapping between a section of a data model to widgets."""

    def __setitem__(self, key: int, value: widgets.QWidget):
        self.addMapping(value, key)

    def __getitem__(self, key: int) -> widgets.QWidget:
        return self.mappedWidgetAt(key)

    def __delitem__(self, key_or_widget: int | widgets.QWidget):
        widget = (
            self.mappedWidgetAt(key_or_widget)
            if isinstance(key_or_widget, int)
            else key_or_widget
        )
        self.removeMapping(widget)

    def _get_map(self):
        maps = super()._get_map()
        maps |= {"submitPolicy": SUBMIT_POLICY, "orientation": constants.ORIENTATION}
        return maps

    def set_orientation(
        self, orientation: constants.OrientationStr | constants.Orientation
    ):
        """Set the orientation of the data widget mapper.

        Args:
            orientation: orientation for the data widget mapper
        """
        self.setOrientation(constants.ORIENTATION.get_enum_value(orientation))

    def get_orientation(self) -> constants.OrientationStr:
        """Return current orientation.

        Returns:
            orientation
        """
        return constants.ORIENTATION.inverse[self.orientation()]

    def set_submit_policy(
        self, policy: SubmitPolicyStr | widgets.QDataWidgetMapper.SubmitPolicy
    ):
        """Set the submit policy of the mapper.

        Args:
            policy: submit_policy for the data widget mapper
        """
        self.setSubmitPolicy(SUBMIT_POLICY.get_enum_value(policy))

    def get_submit_policy(self) -> SubmitPolicyStr:
        """Return current submit policy.

        Returns:
            submit policy
        """
        return SUBMIT_POLICY.inverse[self.submitPolicy()]

    def add_mapping(
        self, widget: widgets.QWidget, section: int, property_name: str | None = None
    ):
        if property_name is None:
            self.addMapping(widget, section)
        else:
            ba = core.QByteArray(property_name.encode())
            self.addMapping(widget, section, ba)

    def get_mapped_property_name(self, widget: widgets.QWidget) -> str:
        return self.mappedPropertyName(widget).data().decode()


if __name__ == "__main__":
    mapper = DataWidgetMapper()



from __future__ import annotations

import asyncio
import functools
import inspect
import itertools
import logging
from typing import TYPE_CHECKING, Any, Optional, get_args, get_type_hints

from prettyqt.qt import QtCore


if TYPE_CHECKING:
    from collections.abc import Callable, Iterable


logger = logging.getLogger(__name__)


def Slot(*args, auto: bool = False, **kwargs):
    if auto:
        return AutoSlot

    def outer_decorator(fn):
        if inspect.iscoroutinefunction(fn):

            @QtCore.Slot(*args)
            @functools.wraps(fn)
            def wrapper(*args, **kwargs):
                return asyncio.create_task(fn(*args, **kwargs))

            return wrapper
        return QtCore.Slot(*args, **kwargs)(fn)

    return outer_decorator


def is_optional(typ) -> bool:
    return typ in (Optional, type(Optional))


# def is_optional(field):
#     return get_origin(field) is Union and type(None) in typing.get_args(field)


def get_optional_args(annotations: Iterable[Any]) -> list[Any]:
    return [get_args(arg)[0] for arg in annotations if is_optional(arg)]


def get_concrete(annotation: Any) -> type | None:
    return getattr(annotation, "__origin__", None)


def get_concretes(annotations: list) -> list:
    ret = []
    for item in annotations:
        if is_optional(item):
            annotations.remove(item)
            continue
        if concrete := get_concrete(item):
            ret.append(concrete)
            annotations.remove(item)
    return ret + annotations


def AutoSlot(func):
    def wrapper(func) -> func:
        anots: dict[str, Any] = func.__annotations__
        return_ = anots.pop("return", None)
        if return_ is Any:
            return_ = "QVariant"
        args = list(anots.values())
        stripped_optionals = get_concretes(get_optional_args(args))
        concretes = get_concretes(args)
        if stripped_optionals:
            required_annotations = tuple(concretes)
            combos = [
                subset
                for i in range(len(stripped_optionals) + 1)
                for subset in itertools.combinations(stripped_optionals, i)
            ]
            for combo in combos:
                func = QtCore.Slot(*combo + required_annotations, result=return_)(func)
            return func
        return QtCore.Slot(*concretes, result=return_)(func)

    return wrapper(func)


class _OverloadedSlotPlaceholder:
    def __init__(self, name, original_name):
        self.name = name
        self.original_name = original_name

    def __repr__(self):
        return f"<placeholder slot {self.name} for {self.original_name}>"


def _build_arguments(
    func: Callable, original_func: Callable | None = None
) -> tuple[list, dict]:
    argspec = inspect.getfullargspec(func)
    args = argspec.args
    annotations = get_type_hints(func)

    slot_args: list[Any] = []
    slot_kwargs: dict[str, Any] = {}

    if "return" in annotations and annotations["return"] is not None:
        slot_kwargs["result"] = annotations["return"]

    if original_func and func.__name__ != original_func.__name__:
        slot_kwargs["name"] = func.__name__

    gap = False

    for index, arg in enumerate(args):
        # Let the first arg pass by unannotated, as it may be the 'self'
        # argument of a function. Static methods can't be used as slots anyways
        if index == 0 and arg not in annotations:
            continue

        if arg in annotations and not gap:
            slot_args.append(annotations[arg])
        elif arg not in annotations:
            gap = True
        else:
            msg = (
                f"Annotations must be in a continuous row - arg before '{arg}' is missing"
            )
            raise TypeError(msg)

    return slot_args, slot_kwargs


def AutoSlotAlternative(func: Callable):
    """A special slot decorator using function annotations to provide the argument types.

    Use like so:

        @slot
        def my_function(self, x: int, y: int) -> int:
            return x ** y

    You can also overload a function by doing this:

        @slot
        def overloaded_func(self, output=None: str):
            print(output)

        @overloaded_func.overload
        def overloaded_func(self, output: int): ...

        # And an overload for calling it with no arguments, so it will fall back
        # on the default argument.
        @overloaded_func.overload
        def overloaded_func(self): ...

    Note: you cannot set different default arguments in the overloads, you may
    only set a function's default arguments in the main declaration.

    The function body on overloads is completely ignored, so you can just use
    ellipses (`...`) or `pass` to fill in the body.
    """
    args, kwargs = _build_arguments(func)
    logger.debug("Auto-slot for %s: args: %s, kwargs: %s", func.__name__, args, kwargs)

    QtCore.Slot(*args, **kwargs)(func)

    def overload(new_func):
        args, kwargs = _build_arguments(new_func, func)

        QtCore.Slot(*args, **kwargs)(func)

        if new_func.__name__ != func.__name__:
            return _OverloadedSlotPlaceholder(new_func.__name__, func.__name__)
        return func

    func.overload = overload  # type: ignore

    return func



from __future__ import annotations

from prettyqt import core


class IdentityProxyModel(core.AbstractProxyModelMixin, core.QIdentityProxyModel):
    """Proxies its source model unmodified."""

    ID = "identity"



from __future__ import annotations

from typing import Literal

from prettyqt import core
from prettyqt.utils import bidict, get_repr


SystemStr = Literal["gregorian", "julian", "milankovic", "jalali", "islamic_civil"]

SYSTEM: bidict[SystemStr, core.QCalendar.System] = bidict(
    gregorian=core.QCalendar.System.Gregorian,
    julian=core.QCalendar.System.Julian,
    milankovic=core.QCalendar.System.Milankovic,
    jalali=core.QCalendar.System.Jalali,
    islamic_civil=core.QCalendar.System.IslamicCivil,
)


class Calendar(core.QCalendar):
    """Describes calendar systems."""

    def __init__(self, system: str | core.QCalendar.System = "gregorian"):
        typ = system if isinstance(system, core.QCalendar.System) else SYSTEM[system]
        super().__init__(typ)

    def __repr__(self):
        return get_repr(self, self.name())

    def __reduce__(self):
        return type(self), (self.name(),)

    # def __bool__(self):
    #     return self.isValid()

    def get_date_from_parts(self, year: int, month: int, day: int) -> core.Date:
        return core.Date(self.dateFromParts(year, month, day))


if __name__ == "__main__":
    cal = Calendar("gregorian")



"""Provides classes for creating and executing state graphs."""

from __future__ import annotations

from prettyqt.qt.QtStateMachine import *  # noqa: F403

from .abstractstate import AbstractState, AbstractStateMixin
from .abstracttransition import AbstractTransition, AbstractTransitionMixin
from .eventtransition import EventTransition
from .finalstate import FinalState
from .historystate import HistoryState
from .signaltransition import SignalTransition
from .state import State
from .statemachine import StateMachine
from prettyqt.qt import QtStateMachine

QT_MODULE = QtStateMachine

__all__ = [
    "AbstractState",
    "AbstractStateMixin",
    "AbstractTransition",
    "AbstractTransitionMixin",
    "EventTransition",
    "FinalState",
    "HistoryState",
    "SignalTransition",
    "State",
    "StateMachine",
]



from __future__ import annotations

import enum
import logging
from typing import TYPE_CHECKING, Any

from prettyqt import constants, core, itemmodels


if TYPE_CHECKING:
    from collections.abc import Iterable, Sequence

    from prettyqt import gui


logger = logging.getLogger(__name__)

SMALL_COL_WIDTH = 120
MEDIUM_COL_WIDTH = 200


class ColumnItem:
    """Determines how an object attribute is shown."""

    model: core.QAbstractItemModel
    name: str = ""
    doc: str = "<no help available>"
    col_visible: bool = True
    width: int | str = SMALL_COL_WIDTH
    line_wrap: gui.textoption.WordWrapModeStr = "none"
    selectable: bool = True
    enabled: bool = True
    editable: bool = False
    checkable: bool = False
    tristate: bool = False

    def __init__(self, model: core.QAbstractItemModel):
        self.model = model

    def get_name(self) -> str:
        return self.name

    def get_flags(self, tree_item):
        flag = constants.NO_FLAGS
        if self.selectable:
            flag |= constants.IS_SELECTABLE
        if self.enabled:
            flag |= constants.IS_ENABLED
        if self.editable:
            flag |= constants.IS_EDITABLE
        if self.checkable:
            flag |= constants.IS_CHECKABLE
        if self.tristate:
            flag |= constants.IS_USER_TRISTATE
        return flag

    def get_data(self, item, role):
        return NotImplemented

    def set_data(self, item, value, role):
        return NotImplemented

    def get_width(self) -> int:
        match self.width:
            case "small":
                return SMALL_COL_WIDTH
            case "medium":
                return MEDIUM_COL_WIDTH
            case int():
                return self.width
            case _:
                raise ValueError(self.width)

    @staticmethod
    def to_checkstate(value: bool):
        return constants.CheckState.Checked if value else constants.CheckState.Unchecked


class ColumnItemModelMixin:
    class ExtraRoles(enum.IntEnum):
        """Addional roles."""

        TreeItemRole = constants.USER_ROLE + 43435
        ColumnItemRole = constants.USER_ROLE + 43436

    def set_columns(self, columns: Sequence[type[ColumnItem]]):
        self._attr_cols = [Col(model=self) for Col in columns]

    def data(
        self,
        index: core.ModelIndex,
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
    ):
        """Return the tree item at the given index and role."""
        if not index.isValid():
            return None

        col = index.column()
        tree_item = self.data_by_index(index)
        col_item = self._attr_cols[col]
        match role:
            case self.ExtraRoles.TreeItemRole:
                return tree_item
            case self.ExtraRoles.ColumnItemRole:
                return col_item
            case _:
                return col_item.get_data(tree_item, role)

    def setData(
        self,
        index: core.ModelIndex,
        value: Any,
        role: constants.ItemDataRole = constants.EDIT_ROLE,
    ) -> bool:
        if not index.isValid():
            return False
        col = index.column()
        tree_item = self.data_by_index(index)
        col_item = self._attr_cols[col]
        col_item.set_data(tree_item, value, role)
        self.dataChanged.emit(index, index)
        return True

    def flags(self, index: core.ModelIndex) -> constants.ItemFlag:
        if not index.isValid():
            return super().flags(index)  # TODO: whats best here?
        col = index.column()
        tree_item = self.data_by_index(index)
        return self._attr_cols[col].get_flags(tree_item)

    def headerData(
        self,
        section: int,
        orientation: constants.Orientation,
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
    ) -> str | None:
        match orientation, role:
            case constants.HORIZONTAL, constants.DISPLAY_ROLE:
                return self._attr_cols[section].get_name()
            case _, _:
                return None


class ColumnItemModel(ColumnItemModelMixin, itemmodels.TreeModel):
    def __init__(
        self,
        obj=None,
        columns: Sequence[type[ColumnItem]] = [],
        mime_type: str | None = None,
        show_root: bool = True,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self._root_item = self.TreeItem(obj=obj)
        self._show_root = show_root
        self.mime_type = mime_type
        self._attr_cols = []
        self.set_columns(columns)
        self.set_root_item(obj)

    def columnCount(self, parent=None):
        return len(self._attr_cols)


class ColumnTableModel(ColumnItemModelMixin, core.AbstractTableModel):
    def __init__(
        self,
        items: Sequence,
        columns: Sequence[ColumnItem],
        mime_type: str | None = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.items = items
        self.mime_type = mime_type
        self._attr_cols = []
        self.set_columns(columns)

    def rowCount(self, parent: core.ModelIndex | None = None) -> int:
        parent = parent or core.ModelIndex()
        return 0 if parent.isValid() else len(self.items)

    def columnCount(self, parent: core.ModelIndex | None = None) -> int:
        parent = parent or core.ModelIndex()
        return 0 if parent.isValid() else len(self._attr_cols)

    def data_by_index(self, index: core.ModelIndex):
        return self.items[index.row()]

    def setData(
        self,
        index: core.ModelIndex,
        value: Any,
        role: constants.ItemDataRole = constants.EDIT_ROLE,
    ):
        if role == constants.USER_ROLE:
            self.items[index.row()] = value
            self.update_row(index.row())
            return True
        return super().setData(index, value, role)

    def removeRows(self, row: int, count: int, parent):
        end_row = row + count - 1
        with self.remove_rows(row, end_row, parent):
            for i in range(end_row, row - 1, -1):
                self.items.pop(i)
        return True

    def dropMimeData(self, mime_data, action, row, column, parent_index):
        if not mime_data.hasFormat(self.mime_type):
            return False
        # Since we only drop in between items, parent_index must be invalid,
        # and we use the row arg to know where the drop took place.
        if parent_index.isValid():
            return False
        indexes = mime_data.get_json_data(self.mime_type)
        pos = row if row < len(self.items) and row != -1 else len(self.items)
        rem_offset = sum(i <= pos for i in indexes)
        new = [self.items[i] for i in indexes]
        with self.change_layout():
            for i in sorted(indexes, reverse=True):
                self.items.pop(i)
            for item in reversed(new):
                self.items.insert(pos - rem_offset, item)
        return False

    def sort(self, ncol: int, order):
        """Sort table by given column number."""
        is_asc = order == constants.ASCENDING
        if sorter := self._attr_cols[ncol].label:
            with self.change_layout():
                self.items.sort(key=sorter, reverse=is_asc)

    def add(self, item: Any, position: int | None = None):
        """Append provided item to the list."""
        self.add_items(items=[item], position=position)
        return item

    def add_items(self, items: Iterable[Any], position: int | None = None):
        """Append a list of items to the list."""
        if position is None:
            position = len(self.items)
        items = list(items)
        with self.insert_rows(position, position + len(items) - 1):
            for i, _ in enumerate(items):
                self.items.insert(i + position, items[i])
            # self.items.extend(items)
        return items

    def remove_items(self, offsets: Iterable[int]):
        for offset in sorted(offsets, reverse=True):
            self.removeRow(offset)


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    test = {}

    class TestColumn(ColumnItem):
        name = "test"
        checkable = True

        def get_data(self, item, role):
            match role:
                case constants.CHECKSTATE_ROLE:
                    return getattr(item, "test", False)
                case constants.DISPLAY_ROLE:
                    return str(item.get_root_path())
                case constants.USER_ROLE:
                    return str(item.get_root_path())

        def set_data(self, item, value, role):
            match role:
                case constants.CHECKSTATE_ROLE:
                    item.test = value

    items = core.StorageInfo.get_mounted_volumes()
    model = ColumnTableModel(items, [TestColumn])
    table = widgets.TableView()
    table.setSortingEnabled(True)
    table.set_model(model)
    table.show()
    app.exec()



from __future__ import annotations

from typing import TYPE_CHECKING, Self

from prettyqt import constants, core


if TYPE_CHECKING:
    import datetime


class DateTime(core.QDateTime):
    """DateTime funcitons."""

    def __repr__(self):
        super_repr = super().__repr__()
        template = str(super_repr).split("(")[1]
        return f"{type(self).__name__}({template}"

    def __str__(self):
        return self.toString("yyyy-MM-dd hh:mm:ss.zzzzzz")

    def __reduce__(self):
        return type(self), (self.date(), self.time(), self.get_timezone())

    def __format__(self, format_spec: str):
        if format_spec in constants.DATE_FORMAT:
            return self.toString(constants.DATE_FORMAT[format_spec])
        return self.toString(format_spec)

    @classmethod
    def from_seconds(cls, seconds: float) -> Self:
        new = cls()
        new.setMSecsSinceEpoch(int(seconds * 1000))
        return new

    def get_value(self) -> datetime.datetime:
        return self.toPython()  # type: ignore

    def get_date(self) -> core.Date:
        return core.Date(self.date())

    def get_time(self) -> core.Time:
        return core.Time(self.time())

    def get_timezone(self) -> core.TimeZone:
        return core.TimeZone(self.timeZone())

    def set_timezone(self, zone: str | core.QTimeZone):
        if isinstance(zone, str):
            self.setTimeZone(core.TimeZone(zone))
        else:
            self.setTimeZone(zone)

    def set_time_spec(self, spec: constants.TimeSpecStr | constants.TimeSpec):
        """Set the time specification.

        Args:
            spec: time specification to use
        """
        self.setTimeSpec(constants.TIME_SPEC[spec])

    def get_time_spec(self) -> constants.TimeSpecStr:
        """Return current time specification.

        Returns:
            time specification
        """
        return constants.TIME_SPEC.inverse[self.timeSpec()]

    def to_format(self, fmt: constants.DateFormatStr):
        return self.toString(constants.DATE_FORMAT[fmt])


if __name__ == "__main__":
    date = DateTime(2000, 11, 11, 1, 1, 1)
    dt = DateTime(date)
    pdt = dt.get_value()



from __future__ import annotations

from prettyqt import positioning
from prettyqt.qt import QtPositioning
from prettyqt.utils import get_repr


class GeoCircle(positioning.GeoShapeMixin, QtPositioning.QGeoCircle):
    def __init__(
        self,
        center_or_other: None
        | (
            QtPositioning.QGeoShape | QtPositioning.QGeoCoordinate | tuple[float, float]
        ) = None,
        radius: float | None = None,
    ):
        if center_or_other is None:
            super().__init__()
        else:
            if radius is None:
                radius = -1
            if isinstance(center_or_other, tuple):
                center_or_other = QtPositioning.QGeoCoordinate(*center_or_other)
            super().__init__(center_or_other, radius)

    def __repr__(self):
        return get_repr(self, self.get_center(), self.radius())

    def get_center(self) -> positioning.GeoCoordinate:
        return positioning.GeoCoordinate(self.center())


if __name__ == "__main__":
    coord = (1, 1)
    circle = GeoCircle(coord)
    print(circle)
    print(repr(circle))
